import functools
import re
import warnings
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Tuple, Union

import onetick.py as otp
from onetick.py.backports import Literal
from onetick.py import types as ott
from onetick.py import utils
from onetick.py.core.column import _Column
from onetick.py.otq import otq
from onetick.py.compatibility import (
    is_ob_virtual_prl_and_show_full_detail_supported,
    is_per_tick_script_boolean_problem,
)
from onetick.py.aggregations._docs import copy_signature

if TYPE_CHECKING:
    from onetick.py.core.source import Source


def inplace_operation(method):
    """Decorator that adds the `inplace` parameter and logic according to this
    flag. inplace=True means that method modifies an object, otherwise it copies
    the object firstly, modifies copy and returns the copy.
    """

    @functools.wraps(method)
    def _inner(self, *args, inplace=False, **kwargs):
        kwargs['inplace'] = inplace
        if inplace:
            method(self, *args, **kwargs)
            return None
        else:
            obj = self.copy()
            return method(obj, *args, **kwargs)

    return _inner


def _columns_names_regex(
    self: 'Source', objs: Tuple[Union[_Column, str]], drop: bool = False
) -> Tuple[List[str], bool]:
    """
    We can't be sure python Source has consistent columns cache, because sinking complex event processors
    can change columns unpredictable, so if user will specify regex as a param, we will pass regex
    as an onetick's param, but pass or delete all matched columns from python Source cache.

    Parameters
    ----------
    objs:
        Tuple of _Columns or string to pass or drop. String can be regex.
    drop: bool
        To drop columns from python schema or not.

    Returns
    -------
    items_to_passthrough:
        Items to pass to Passthrough as `field` parameter.
    regex:
        Value to pass to Passthrough as `use_regex` parameter.
    """
    def is_regex_string(obj):
        return isinstance(obj, str) and any(c in r"*+?\:[]{}()^$" for c in obj)

    # if any object from the list is a regexp, then all other objects will be treated like regexps too
    regex = any(is_regex_string(obj) for obj in objs)

    items_to_passthrough = []
    names_of_columns = []
    for obj in objs:
        if not isinstance(obj, (str, _Column)):
            raise TypeError(f"It is not supported to select or delete item '{obj}' of type '{type(obj)}'")
        if regex:
            # if column object or non-regex string is specified in regex mode,
            # then we assume that the user requested that exact column to be dropped
            if isinstance(obj, _Column):
                names_of_columns.append(obj.name)
                obj = f'^{obj.name}$'
            elif not is_regex_string(obj):
                names_of_columns.append(obj)
                obj = f'^{obj}$'
            else:
                names_of_columns.extend(col for col in self.columns() if re.search(obj, col))
            items_to_passthrough.append(obj)
        else:
            name = obj.name if isinstance(obj, _Column) else obj
            items_to_passthrough.append(name)
            names_of_columns.append(name)

    # remove duplications and meta_fields
    names_of_columns: set[str] = set(names_of_columns) - set(self.__class__.meta_fields)  # type: ignore[no-redef]
    # TODO: we definitely have the same logic of checking columns somewhere else too, need to refactor
    for item in names_of_columns:
        if item not in self.__dict__ or not isinstance(self.__dict__[item], _Column):
            raise AttributeError(f"There is no '{item}' column")
    if drop:
        for item in names_of_columns:
            del self.__dict__[item]
    return items_to_passthrough, regex


@inplace_operation
def pause(self: 'Source', delay, busy_waiting=False, where=None, inplace=False) -> Optional['Source']:
    """
    Pauses processing of each tick for number of milliseconds specified via ``delay`` expression.

    Parameters
    ----------
    delay: int or :py:class:`onetick.py.Operation`
        Integer number or OneTick expression used to calculate the delay.
        Delay is in milliseconds.
        Note that number can't be negative.
    busy_waiting: bool
        If True then delay is done via busy loop (consuming CPU time).
    where: :py:class:`onetick.py.Operation`
        Expression to select ticks for which processing will be paused.
        By default, all ticks are selected.
    inplace: bool
        The flag controls whether operation should be applied inplace or not.
        If ``inplace=True``, then it returns nothing.
        Otherwise method returns a new modified object.

    See also
    --------
    **PAUSE** OneTick event processor

    Returns
    -------
    :class:`Source` or ``None``

    Examples
    --------

    Pause every tick for 100 milliseconds (0.1 second).

    >>> data = otp.Tick(A=1)
    >>> data = data.pause(100)

    Set the ``delay`` with expression and use ``where`` parameter to pause conditionally:

    >>> data = otp.Ticks(A=[-1, 1, 2])
    >>> data = data.pause(data['A'] * 100, where=data['A'] > 0)
    """
    if not isinstance(delay, int):
        delay = str(delay)
    elif delay < 0:
        raise ValueError("Parameter 'delay' can't be negative")
    where = '' if where is None else str(where)
    self.sink(
        otq.Pause(
            delay=delay,
            busy_waiting=busy_waiting,
            where=where,
        )
    )
    return self


@inplace_operation
def insert_tick(
    self: 'Source',
    fields=None,
    where=None,
    preserve_input_ticks=True,
    num_ticks_to_insert=1,
    insert_before=True,
    inplace=False,
) -> Optional['Source']:
    """
    Insert tick.

    Parameters
    ----------
    fields: dict of str to :py:class:`onetick.py.Operation`
        Mapping of field names to some expressions or values.
        These fields in inserted ticks will be set to corresponding values or results of expressions.
        If field is presented in input tick, but not set in ``fields`` dict,
        then the value of the field will be copied from input tick to inserted tick.
        If parameter ``fields`` is not set at all, then values for inserted ticks' fields
        will be default values for fields' types from input ticks (0 for integers etc.).
    where: :py:class:`onetick.py.Operation`
        Expression to select ticks near which the new ticks will be inserted.
        By default, all ticks are selected.
    preserve_input_ticks: bool
        A switch controlling whether input ticks have to be preserved in output time series or not.
        While the former case results in fields of input ticks to be present in the output time series
        together with those defined by the ``fields`` parameter,
        the latter case results in only defined fields to be present.
        If a field of the input time series is defined in the ``fields`` parameter,
        the defined value takes precedence.
    num_ticks_to_insert: int
        Number of ticks to insert.
    insert_before: bool
        Insert tick before each input tick or after.
    inplace: bool
        The flag controls whether operation should be applied inplace or not.
        If ``inplace=True``, then it returns nothing.
        Otherwise method returns a new modified object.

    See also
    --------
    **INSERT_TICK** OneTick event processor

    Returns
    -------
    :class:`Source` or ``None``

    Examples
    --------

    Insert tick before each tick with default type values.

    >>> data = otp.Tick(A=1)
    >>> data = data.insert_tick()
    >>> otp.run(data)
            Time  A
    0 2003-12-01  0
    1 2003-12-01  1

    Insert tick before each tick with field `A` copied from input tick
    and field `B` set to specified value.

    >>> data = otp.Tick(A=1)
    >>> data = data.insert_tick(fields={'B': 'b'})
    >>> otp.run(data)
            Time  B  A
    0 2003-12-01  b  1
    1 2003-12-01     1

    Insert two ticks only after first tick.

    >>> data = otp.Ticks(A=[1, 2, 3])
    >>> data = data.insert_tick(where=data['A'] == 1,
    ...                         insert_before=False,
    ...                         num_ticks_to_insert=2)
    >>> otp.run(data)
                         Time  A
    0 2003-12-01 00:00:00.000  1
    1 2003-12-01 00:00:00.000  0
    2 2003-12-01 00:00:00.000  0
    3 2003-12-01 00:00:00.001  2
    4 2003-12-01 00:00:00.002  3
    """
    if not isinstance(num_ticks_to_insert, int) or num_ticks_to_insert <= 0:
        raise ValueError("Parameter 'num_ticks_to_insert' must be a positive integer")
    if not preserve_input_ticks and not fields:
        raise ValueError("Parameter 'fields' must be set if 'preserve_input_ticks' is False")
    where = '' if where is None else str(where)

    fields = fields or {}
    update_schema = {}
    for field, value in fields.items():
        dtype = ott.get_object_type(value)
        if field not in self.schema:
            update_schema[field] = dtype
        elif dtype is not self.schema[field]:
            raise ValueError(f"Incompatible types for field '{field}': {self.schema[field]} --> {dtype}")
        dtype = ott.type2str(dtype)
        if isinstance(value, type):
            value = ott.default_by_type(value)
        value = ott.value2str(value)
        fields[field] = (dtype, value)
    fields = ','.join(
        f'{field} {dtype}={value}' if value else f'{field} {dtype}' for field, (dtype, value) in fields.items()
    )

    self.sink(
        otq.InsertTick(
            fields=fields,
            where=where,
            preserve_input_ticks=preserve_input_ticks,
            num_ticks_to_insert=num_ticks_to_insert,
            insert_before=insert_before,
        )
    )
    if preserve_input_ticks:
        self.table(inplace=True, strict=False, **update_schema)
    else:
        self.table(inplace=True, strict=True, **update_schema)
    return self


@inplace_operation
def insert_at_end(
    self: 'Source',
    *,
    propagate_ticks: bool = True,
    delimiter_name: str = 'AT_END',
    inplace: bool = False,
) -> Optional['Source']:
    """
    This function adds a field ``delimiter_name``,
    which is set to zero for all inbound ticks
    and set to 1 for an additional tick that is generated when the data ends.

    The timestamp of the additional tick is set to the query end time.
    The values of all fields from the input schema of additional tick are set to default values for each type.

    Parameters
    ----------
    propagate_ticks: bool
        If True (default) this function returns all input ticks and an additionally generated tick,
        otherwise it returns only the last generated tick.
    delimiter_name: str
        The name of the delimiter field to add.
    inplace: bool
        The flag controls whether operation should be applied inplace or not.
        If ``inplace=True``, then it returns nothing.
        Otherwise method returns a new modified object.

    See also
    --------
    **INSERT_AT_END** OneTick event processor

    Returns
    -------
    :class:`Source` or ``None``

    Examples
    --------
    Insert tick at the end of the stream:

    >>> data = otp.Ticks(A=[1, 2, 3])
    >>> data = data.insert_at_end()
    >>> otp.run(data)
                         Time  A  AT_END
    0 2003-12-01 00:00:00.000  1       0
    1 2003-12-01 00:00:00.001  2       0
    2 2003-12-01 00:00:00.002  3       0
    3 2003-12-04 00:00:00.000  0       1

    The name of the added field can be changed:

    >>> data = otp.Ticks(A=[1, 2, 3])
    >>> data = data.insert_at_end(delimiter_name='LAST_TICK')
    >>> otp.run(data)
                         Time  A  LAST_TICK
    0 2003-12-01 00:00:00.000  1          0
    1 2003-12-01 00:00:00.001  2          0
    2 2003-12-01 00:00:00.002  3          0
    3 2003-12-04 00:00:00.000  0          1

    If parameter ``propagate_ticks`` is set to False, then only the last tick is returned:

    >>> data = otp.Ticks(A=[1, 2, 3])
    >>> data = data.insert_at_end(propagate_ticks=False)
    >>> otp.run(data)
            Time  A  AT_END
    0 2003-12-04  0       1
    """
    if not hasattr(otq, 'InsertAtEnd'):
        raise RuntimeError("Function insert_at_end() is not available on this OneTick API version")
    if delimiter_name in self.schema:
        raise ValueError(f"Field '{delimiter_name}' is already in schema")

    self.sink(
        otq.InsertAtEnd(
            propagate_ticks=propagate_ticks,
            delimiter_name=delimiter_name,
        )
    )
    self.schema.update(**{delimiter_name: int})
    return self


@inplace_operation
def transpose(
    self: 'Source',
    inplace: bool = False,
    direction: Literal['rows', 'columns'] = 'rows',
    n: Optional[int] = None,
) -> Optional['Source']:
    """
    Data transposing.
    The main idea is joining many ticks into one or splitting one tick to many.

    Parameters
    ----------
    inplace: bool, default=False
        if `True` method will modify current object,
        otherwise it will return modified copy of the object.
    direction: 'rows', 'columns', default='rows'
        - `rows`: join certain input ticks (depending on other parameters) with preceding ones.
            Fields of each tick will be added to the output tick and their names will be suffixed
            with **_K** where **K** is the positional number of tick (starting from 1) in reverse order.
            So fields of current tick will be suffixed with **_1**, fields of previous tick will be
            suffixed with **_2** and so on.
        - `columns`: the operation is opposite to `rows`. It splits each input tick to several
            output ticks. Each input tick must have fields with names suffixed with **_K**
            where **K** is the positional number of tick (starting from 1) in reverse order.
    n: Optional[int], default=None
        must be specified only if ``direction`` is 'rows'.
        Joins every **n** number of ticks with **n-1** preceding ticks.

    Returns
    -------
    If ``inplace`` parameter is `True` method will return `None`,
    otherwise it will return modified copy of the object.

    See also
    --------
    **TRANSPOSE** OneTick event processor

    Examples
    --------
    Merging two ticks into one.

    >>> data = otp.Ticks(dict(A=[1, 2],
    ...                       B=[3, 4]))
    >>> data = data.transpose(direction='rows', n=2) # OTdirective: skip-snippet:;
    >>> otp.run(data)
                         Time             TIMESTAMP_1  A_1  B_1 TIMESTAMP_2  A_2  B_2
    0 2003-12-01 00:00:00.001 2003-12-01 00:00:00.001    2    4  2003-12-01    1    3

    And splitting them back into two.

    >>> data = data.transpose(direction='columns') # OTdirective: skip-snippet:;
    >>> otp.run(data)
                         Time  A  B
    0 2003-12-01 00:00:00.000  1  3
    1 2003-12-01 00:00:00.001  2  4
    """

    direction_map = {'rows': 'ROWS_TO_COLUMNS', 'columns': 'COLUMNS_TO_ROWS'}
    n: Union[str, int] = '' if n is None else n  # type: ignore[no-redef]

    self.sink(otq.Transpose(direction=direction_map[direction], key_constraint_values=n))
    # TODO: we should change source's schema after transposing
    return self


def cache(
    self: 'Source',
    cache_name: str,
    delete_if_exists: bool = True,
    inheritability: bool = True,
    otq_params: Union[dict, None] = None,
    time_granularity: int = 0,
    time_granularity_units: Optional[str] = None,
    timezone: str = "",
    time_intervals_to_cache: Optional[List[tuple]] = None,
    allow_delete_to_everyone: bool = False,
    allow_update_to_everyone: bool = False,
    allow_search_to_everyone: bool = True,
    cache_expiration_interval: int = 0,
    start: Optional[ott.datetime] = None,
    end: Optional[ott.datetime] = None,
    read_mode: str = "automatic",
    update_cache: bool = True,
    tick_type: str = "ANY",
    symbol: Optional[str] = None,
    db: Optional[str] = None,
) -> 'Source':
    """
    Create cache from query and :py:class:`onetick.py.ReadCache` for created cache.

    Cache will be created only for current session.

    By default, if cache with specified name exists, it will be deleted and recreated.
    You can change this behaviour via ``delete_if_exists`` parameter.

    Parameters
    ----------
    cache_name: str
        Name of the cache to be deleted.
    delete_if_exists: bool
        If set to ``True`` (default), a check will be made to detect the existence of a cache
        with the specified name. Cache will be deleted and recreated only if it exists.
        If set to ``False``, if cache exists it won't be deleted and recreated.
    inheritability: bool
        Indicates whether results can be obtained by combining time intervals that were cached with intervals
        freshly computed to obtain results for larger intervals.
    otq_params: dict
        OTQ params of the query to be cached.
    time_granularity: int
        Value N for seconds/days/months granularity means that start and end time of the query have to be on N
        second/day/month boundaries relative to start of the day/month/year.
        This doesn't affect the frequency of data within the cache, just the start and end dates.
    time_granularity_units: str, None
        Units used in ``time_granularity`` parameter. Possible values: 'none', 'days', 'months', 'seconds' or None.
    timezone: str
        Timezone of the query to be cached.
    time_intervals_to_cache: List[tuple]
        List of tuples with start and end times in ``[(<start_time_1>, <end_time_1>), ...]`` format,
        where ``<start_time>`` and ``<end_time>`` should be one of these:

        * string in ``YYYYMMDDhhmmss[.msec]`` format.
        * :py:class:`datetime.datetime`
        * :py:class:`onetick.py.types.datetime`

        If specified only these time intervals can be cached. Ignored if ``inheritability=True``.
        If you try to make a query outside defined interval, error will be raised.
    allow_delete_to_everyone: bool
        When set to ``True`` everyone is allowed to delete the cache.
    allow_update_to_everyone: bool
        When set to ``True`` everyone is allowed to update the cache.
    allow_search_to_everyone: bool
        When set to ``True`` everyone is allowed to read the cached data.
    cache_expiration_interval: int
        If set to a non-zero value determines the periodicity of cache clearing, in seconds.
        The cache will be cleared every X seconds, triggering new query executions when data is requested.
    start: :py:class:`otp.datetime <onetick.py.datetime>`
        Start time for cache query. By default, the start time of the query will be used.
    end: :py:class:`otp.datetime <onetick.py.datetime>`
        End time for cache query. By default, the end time of the query will be used.
    read_mode: str
        Mode of querying cache. One of these:

        * ``cache_only`` - only cached results are returned and queries are not performed.
        * ``query_only`` - the query is run irrespective of whether the result is already available in the cache.
        * ``automatic`` (default) - perform the query if the data is not found in the cache.
    update_cache: bool
        If set to ``True``, updates the cached data if ``read_mode=query_only`` or if ``read_mode=automatic`` and
        the result data not found in the cache. Otherwise, the cache remains unchanged.
    tick_type: str
        Tick type.
    symbol: str, list of str, list of otq.Symbol, :py:class:`onetick.py.Source`, :pandas:`pandas.DataFrame`, optional
        ``symbols`` parameter of ``otp.run()``.
    db: str
        Database.

    See also
    --------
    | :py:func:`onetick.py.create_cache`
    | :py:class:`onetick.py.ReadCache`

    Examples
    --------
    Simple usage

    >>> src = otp.DataSource("COMMON", tick_type="TRD", symbols="AAPL")
    >>> data = src.cache(
    ...    cache_name="some_cache",
    ...    tick_type="TRD", symbol="SYM", db="LOCAL",
    ... )
    >>> df = otp.run(data)  # doctest: +SKIP
    """
    from onetick.py.cache import create_cache, delete_cache, modify_cache_config
    from onetick.py.sources import ReadCache

    cache_exists = True

    if delete_if_exists:
        try:
            modify_cache_config(cache_name, "TEST_PARAM", "TEST_VALUE")
        except Exception as exc:
            if "There is no cache" in str(exc):
                cache_exists = False

    if cache_exists and delete_if_exists:
        delete_cache(cache_name)
        cache_exists = False

    if not cache_exists:
        create_cache(
            cache_name=cache_name,
            query=self,
            inheritability=inheritability,
            otq_params=otq_params,
            time_granularity=time_granularity,
            time_granularity_units=time_granularity_units,
            timezone=timezone,
            time_intervals_to_cache=time_intervals_to_cache,
            allow_delete_to_everyone=allow_delete_to_everyone,
            allow_update_to_everyone=allow_update_to_everyone,
            allow_search_to_everyone=allow_search_to_everyone,
            cache_expiration_interval=cache_expiration_interval,
            tick_type=tick_type,
            symbol=symbol,
            db=db,
        )

    return ReadCache(
        cache_name=cache_name,
        start=start if start is not None else utils.adaptive,
        end=end if end is not None else utils.adaptive,
        read_mode=read_mode,
        update_cache=update_cache,
        tick_type=tick_type,
        symbol=symbol if symbol is not None else utils.adaptive,
        db=db if db is not None else utils.adaptive_to_default,
    )


@inplace_operation
def pnl_realized(
    self: 'Source',
    computation_method: str = 'fifo',
    output_field_name: str = 'PNL_REALIZED',
    size_field: Union[str, _Column] = 'SIZE',
    price_field: Union[str, _Column] = 'PRICE',
    buy_sell_flag_field: Union[str, _Column] = 'BUY_SELL_FLAG',
    inplace=False,
) -> Optional['Source']:
    """
    Computes the realized Profit and Loss (**PNL**) on each tick and is applicable to both scenarios,
    whether selling after buying or buying after selling.

    Parameters
    ----------
    computation_method: str
        This parameter determines the approach used to calculate the realized Profit and Loss (**PnL**).

        Possible options are:

        * ``fifo`` - Stands for 'First-In-First-Out,' is used to calculate Profit and Loss (**PnL**)
          based on the principle that the first trading positions bought are the first ones to be sold,
          or conversely, the first trading positions sold are the first ones to be bought.

    output_field_name: str
        This parameter defines the name of the output field.

        Default: **PNL_REALIZED**.

    size_field: str, :py:class:`otp.Column <onetick.py.Column>`
        The name of the field with size, default is **SIZE**.
    price_field: str, :py:class:`otp.Column <onetick.py.Column>`
        The name of the field with price, default is **PRICE**.
    buy_sell_flag_field: str, :py:class:`otp.Column <onetick.py.Column>`
        The name of the field with buy/sell flag, default is **BUY_SELL_FLAG**.
        If the type of this field is string, then possible values are 'B' or 'b' for buy and 'S' or 's' for sell.
        If the type of this field is integer, then possible values are 0 for buy and 1 for sell.


    See also
    --------
    **PNL_REALIZED** OneTick event processor

    Examples
    --------
    Let's generate some data:

    >>> trades = otp.Ticks(
    ...     PRICE=[1.0, 2.0, 3.0, 2.5, 4.0, 5.0, 6.0, 7.0, 3.0, 4.0, 1.0],
    ...     SIZE=[700, 20, 570, 600, 100, 100, 100, 100, 150, 10, 100],
    ...     SELL_FLAG=[0, 0, 0, 1, 1, 1, 1, 1, 0, 0, 1],
    ...     SIDE=['B', 'B', 'B', 'S', 'S', 'S', 'S', 'S', 'B', 'B', 'S'],
    ... )
    >>> otp.run(trades)
                          Time  PRICE  SIZE  SELL_FLAG SIDE
    0  2003-12-01 00:00:00.000    1.0   700          0    B
    1  2003-12-01 00:00:00.001    2.0    20          0    B
    2  2003-12-01 00:00:00.002    3.0   570          0    B
    3  2003-12-01 00:00:00.003    2.5   600          1    S
    4  2003-12-01 00:00:00.004    4.0   100          1    S
    5  2003-12-01 00:00:00.005    5.0   100          1    S
    6  2003-12-01 00:00:00.006    6.0   100          1    S
    7  2003-12-01 00:00:00.007    7.0   100          1    S
    8  2003-12-01 00:00:00.008    3.0   150          0    B
    9  2003-12-01 00:00:00.009    4.0    10          0    B
    10 2003-12-01 00:00:00.010    1.0   100          1    S

    And then calculate profit and loss metric for it.

    First let's use string ``buy_sell_flag_field`` field:

    >>> data = trades.pnl_realized(buy_sell_flag_field='SIDE')  # doctest: +SKIP
    >>> otp.run(data)[['Time', 'PRICE', 'SIZE', 'SIDE', 'PNL_REALIZED']]  # doctest: +SKIP
                          Time  PRICE  SIZE  SIDE  PNL_REALIZED
    0  2003-12-01 00:00:00.000    1.0   700     B           0.0
    1  2003-12-01 00:00:00.001    2.0    20     B           0.0
    2  2003-12-01 00:00:00.002    3.0   570     B           0.0
    3  2003-12-01 00:00:00.003    2.5   600     S         900.0
    4  2003-12-01 00:00:00.004    4.0   100     S         300.0
    5  2003-12-01 00:00:00.005    5.0   100     S         220.0
    6  2003-12-01 00:00:00.006    6.0   100     S         300.0
    7  2003-12-01 00:00:00.007    7.0   100     S         400.0
    8  2003-12-01 00:00:00.008    3.0   150     B           0.0
    9  2003-12-01 00:00:00.009    4.0    10     B           0.0
    10 2003-12-01 00:00:00.010    1.0   100     S        -200.0

    We can get the same result using integer ``buy_sell_flag_field`` field:

    >>> data = trades.pnl_realized(buy_sell_flag_field='SELL_FLAG')  # doctest: +SKIP
    >>> otp.run(data)[['Time', 'PRICE', 'SIZE', 'SELL_FLAG', 'PNL_REALIZED']]  # doctest: +SKIP
                          Time  PRICE  SIZE  SELL_FLAG  PNL_REALIZED
    0  2003-12-01 00:00:00.000    1.0   700          0           0.0
    1  2003-12-01 00:00:00.001    2.0    20          0           0.0
    2  2003-12-01 00:00:00.002    3.0   570          0           0.0
    3  2003-12-01 00:00:00.003    2.5   600          1         900.0
    4  2003-12-01 00:00:00.004    4.0   100          1         300.0
    5  2003-12-01 00:00:00.005    5.0   100          1         220.0
    6  2003-12-01 00:00:00.006    6.0   100          1         300.0
    7  2003-12-01 00:00:00.007    7.0   100          1         400.0
    8  2003-12-01 00:00:00.008    3.0   150          0           0.0
    9  2003-12-01 00:00:00.009    4.0    10          0           0.0
    10 2003-12-01 00:00:00.010    1.0   100          1        -200.0
    """
    if computation_method not in ['fifo']:
        raise ValueError(
            f"Parameter 'computation_method' has incorrect value: '{computation_method}', "
            f"should be one of these: 'fifo'."
        )

    for field in (size_field, price_field, buy_sell_flag_field):
        if str(field) not in self.schema:
            raise ValueError(f'Field {field} is not in schema')

    # PNL_REALIZED doesn't support choosing input fields, so we workaround by renaming fields
    restore_fields = {}
    added_fields = []
    for default_field, field in (('SIZE', size_field), ('PRICE', price_field), ('BUY_SELL_FLAG', buy_sell_flag_field)):
        if field != default_field:
            if default_field in self.schema:
                tmp_field_name = f'__TMP__{field}__TMP__'
                self[tmp_field_name] = self[default_field]
                restore_fields[default_field] = self[tmp_field_name]
            self[default_field] = self[field]
            added_fields.append(default_field)

    if output_field_name in self.schema:
        raise ValueError(f'Field {output_field_name} is already in schema')

    computation_method = computation_method.upper()
    self.sink(otq.PnlRealized(computation_method=computation_method, output_field_name=output_field_name))
    self.schema[output_field_name] = float

    if added_fields:
        self.drop(added_fields, inplace=True)
    if restore_fields:
        # restore fields from temporary fields and delete temporary fields
        self.add_fields(restore_fields)
        self.drop(list(restore_fields.values()), inplace=True)

    return self


@inplace_operation
def execute(self: 'Source', *operations, inplace=False) -> Optional['Source']:
    """
    Execute operations without returning their values.
    Some operations in onetick.py can be used to modify the state of some object
    (tick sequences mostly) and in that case user may not want to save the result of the
    operation to column.

    Parameters
    ----------
    operations : list of :py:class:`~onetick.py.Operation`
        operations to execute.
    inplace : bool
        The flag controls whether operation should be applied inplace or not.
        If ``inplace=True``, then it returns nothing. Otherwise method returns a new modified
        object.

    Returns
    -------
    :class:`Source` or ``None``

    See Also
    --------
    **EXECUTE_EXPRESSIONS** OneTick event processor

    Examples
    --------
    >>> data = otp.Tick(A=1)
    >>> data.state_vars['SET'] = otp.state.tick_set('oldest', 'A')
    >>> data = data.execute(data.state_vars['SET'].erase(A=1))
    """
    if not operations:
        raise ValueError('At least one operation must be specified in execute() method')
    op_str = ';'.join(map(str, operations))
    self.sink(otq.ExecuteExpressions(op_str))
    return self


@inplace_operation
def fillna(self: 'Source', value=None, columns=None, inplace=False) -> Optional['Source']:
    """
    Replace NaN values in floating point type fields with the ``value``.

    Parameters
    ----------
    value : float, :py:class:`~onetick.py.Operation`
        The value to replace NaN.
        If not specified then the value from the previous tick will be used.
    columns: list
        List of strings with column names or :py:class:`~onetick.py.Column` objects.
        Only the values in specified columns will be replaced.
        By default the values in all floating point type fields will be replaced.
    inplace : bool
        The flag controls whether operation should be applied inplace or not.
        If ``inplace=True``, then it returns nothing. Otherwise method returns a new modified
        object.

    Returns
    -------
    :class:`Source` or ``None``

    See Also
    --------
    :py:meth:`onetick.py.Operation.fillna`

    Examples
    --------

    By default, the value of the previous tick is used as a value to replace NaN
    (for the first tick the previous value do not exist, so it will still be NaN):

    >>> data = otp.Ticks({'A': [0, 1, 2, 3], 'B': [otp.nan, 2.2, otp.nan, 3.3]})
    >>> data = data.fillna()
    >>> otp.run(data)
                         Time  A    B
    0 2003-12-01 00:00:00.000  0  NaN
    1 2003-12-01 00:00:00.001  1  2.2
    2 2003-12-01 00:00:00.002  2  2.2
    3 2003-12-01 00:00:00.003  3  3.3

    The value can also be a constant:

    >>> data = otp.Ticks({'A': [0, 1, 2, 3], 'B': [otp.nan, 2.2, otp.nan, 3.3]})
    >>> data = data.fillna(777)
    >>> otp.run(data)
                         Time  A     B
    0 2003-12-01 00:00:00.000  0  777.0
    1 2003-12-01 00:00:00.001  1    2.2
    2 2003-12-01 00:00:00.002  2  777.0
    3 2003-12-01 00:00:00.003  3    3.3

    :py:class:`~onetick.py.Operation` objects can also be used as a ``value``:

    >>> data = otp.Ticks({'A': [0, 1, 2, 3], 'B': [otp.nan, 2.2, otp.nan, 3.3]})
    >>> data = data.fillna(data['A'])
    >>> otp.run(data)
                         Time  A     B
    0 2003-12-01 00:00:00.000  0    0.0
    1 2003-12-01 00:00:00.001  1    2.2
    2 2003-12-01 00:00:00.002  2    2.0
    3 2003-12-01 00:00:00.003  3    3.3

    Parameter ``columns`` can be used to specify the columns where values will be replaced:

    .. testcode::
       :skipif: is_per_tick_script_boolean_problem()

       data = otp.Ticks({'A': [0, 1, 2, 3], 'B': [otp.nan, 2.2, otp.nan, 3.3], 'C': [otp.nan, 2.2, otp.nan, 3.3]})
       data = data.fillna(columns=['B'])
       df = otp.run(data)
       print(df)

    .. testoutput::

                            Time  A    B    C
       0 2003-12-01 00:00:00.000  0  NaN  NaN
       1 2003-12-01 00:00:00.001  1  2.2  2.2
       2 2003-12-01 00:00:00.002  2  2.2  NaN
       3 2003-12-01 00:00:00.003  3  3.3  3.3
    """
    if columns:
        for column in columns:
            if column not in self.schema:
                raise ValueError(f"Column '{column}' is not in schema")
            if self.schema[column] is not float:
                raise TypeError(f"Column '{column}' doesn't have float type")
        columns = list(map(str, columns))
    no_columns = columns is None

    if value is not None:
        if isinstance(value, int):
            value = float(value)
        if isinstance(value, otp.Operation) and value.dtype is int:
            value = value.astype(float)
        dtype = ott.get_object_type(value)
        if dtype is not float:
            raise ValueError(f"The type of parameter 'value' must be float, not {dtype}")

        def fun(tick):
            for field in otp.tick_descriptor_fields():
                if (field.get_type() == 'double'
                        and tick.get_double_value(field.get_name()) == otp.nan
                        and (no_columns or field.get_name().isin(*columns))):
                    tick.set_double_value(field.get_name(), value)
    else:
        # deque may have already been created by previous call of fillna() method
        if '__OTP_FILLNA_DEQUE__' not in self.state_vars:
            self.state_vars['__OTP_FILLNA_DEQUE__'] = otp.state.tick_deque(scope='branch')

        def fun(tick):
            i = otp.static(0)
            prev_tick = otp.tick_deque_tick()
            # first tick doesn't have the previous tick, so skipping it
            if i > 0:
                # get the previous tick from the deque
                tick.state_vars['__OTP_FILLNA_DEQUE__'].get_tick(0, prev_tick)
                # replace value in all double fields with the value of the previous tick
                for field in otp.tick_descriptor_fields():
                    if (field.get_type() == 'double'
                            and tick.get_double_value(field.get_name()) == otp.nan
                            and (no_columns or field.get_name().isin(*columns))):
                        tick.set_double_value(field.get_name(),
                                              prev_tick.get_double_value(field.get_name()))
            # clear the deque and add the current tick to it
            tick.state_vars['__OTP_FILLNA_DEQUE__'].clear()
            tick.state_vars['__OTP_FILLNA_DEQUE__'].push_back(tick)
            i = i + 1

    return self.script(fun, inplace=inplace)


@inplace_operation
def mkt_activity(self: 'Source', calendar_name=None, inplace=False) -> Optional['Source']:
    """
    Adds a string field named **MKT_ACTIVITY** to each tick in the input tick stream.

    The value of this field is set to the union of session flags that apply for the security at the time of the tick,
    as specified in the calendar sections of the reference database (see Reference Database Guide).

    Session flags may differ between databases,
    but the following letters have reserved meaning: L - half day, H - holiday, W - weekend, R - regular.

    The calendar can either be specified explicitly by name
    (the *CALENDAR* sections of the reference database are assumed to contain a calendar with such a name),
    or default to the security- or exchange-level calendars for the queried symbol
    (the *SYMBOL_CALENDAR* or *EXCH_CALENDAR* sections of the reference database).

    The latter case requires a non-zero symbol date to be specified for queried symbols
    (see parameter ``symbol_date`` in :py:func:`otp.run <onetick.py.run>`).

    Parameters
    ----------
    calendar_name : str, :py:class:`~onetick.py.Column`
        The calendar name to choose for the respective calendar from the *CALENDAR* sections of the reference database.
        It can be a string constant or the name of the field with per-tick calendar name.

        When this parameter is not specified,
        default security- or exchange-level calendars configured for the queried database and symbol are used
        (but symbol date must be specified in this case).
    inplace : bool
        The flag controls whether operation should be applied inplace or not.
        If ``inplace=True``, then it returns nothing. Otherwise method returns a new modified
        object.

    Note
    ----
    When applying :py:meth:`mkt_activity` to aggregated data,
    please take into account that session flags may change during the aggregation bucket.
    The field **MKT_ACTIVITY**, in this case, will represent the session flags at the time assigned to the bucket,
    which may be different from the session flags at some times during the bucket.

    Returns
    -------
    :class:`Source` or ``None``

    See Also
    --------
    **MKT_ACTIVITY** OneTick event processor

    Examples
    --------

    By default, security- or exchange-level calendars configured for the queried database and symbol are used
    (but symbol date must be specified in this case):

    >>> data = otp.DataSource(...)  # doctest: +SKIP
    >>> data = data.mkt_activity()  # doctest: +SKIP
    >>> otp.run(data, date=otp.date(2022, 1, 1), symbol_date=otp.date(2022, 1, 1))  # doctest: +SKIP

    Otherwise, parameter ``calendar_name`` must be specified:

    >>> data = otp.DataSource(...)  # doctest: +SKIP
    >>> data = data.mkt_activity(calendar_name='WNY')  # doctest: +SKIP
    >>> otp.run(data, date=otp.date(2022, 1, 1))  # doctest: +SKIP

    Parameter ``calendar_name`` can also be specified as a column.
    In this case calendar name can be different for each tick:

    >>> data = otp.DataSource(...)  # doctest: +SKIP
    >>> data = data.mkt_activity(calendar_name=data['CALENDAR_NAME'])  # doctest: +SKIP
    >>> otp.run(data, date=otp.date(2022, 1, 1))  # doctest: +SKIP

    In this example you can see how market activity status is changing during the days.
    We are getting first and last tick of the group each time the type of market activity is changed.
    You can see regular trades (R) from 9:30 to 16:00, and a holiday (H) on 2018-02-07.

    >>> data = otp.DataSource('TRAIN_A_PRL_TRD', tick_type='TRD', symbols='MSFT')  # doctest: +SKIP
    >>> data = data.mkt_activity('FRED')  # doctest: +SKIP
    >>> data = data[['PRICE', 'SIZE', 'MKT_ACTIVITY']]  # doctest: +SKIP
    >>> first = data.first(1, bucket_interval=(data['MKT_ACTIVITY'] != data['MKT_ACTIVITY'][-1]))  # doctest: +SKIP
    >>> last = data.last(1, bucket_interval=(data['MKT_ACTIVITY'] != data['MKT_ACTIVITY'][-1]))  # doctest: +SKIP
    >>> data = otp.merge([first, last])  # doctest: +SKIP
    >>> df = otp.run(data,  # doctest: +SKIP
    ...              start=otp.dt(2018, 2, 1), end=otp.dt(2018, 2, 9),
    ...              symbol_date=otp.dt(2018, 2, 1), timezone='EST5EDT')
    >>> df[['Time', 'MKT_ACTIVITY']]  # doctest: +SKIP
                          Time MKT_ACTIVITY
    0  2018-02-01 01:31:44.466
    1  2018-02-01 09:29:59.996
    2  2018-02-01 09:30:00.225            R
    3  2018-02-01 15:59:58.857            R
    4  2018-02-01 16:00:01.858
    5  2018-02-02 09:29:50.366
    6  2018-02-02 09:30:01.847            R
    7  2018-02-02 15:59:59.829            R
    8  2018-02-02 16:00:01.782
    9  2018-02-05 09:29:43.084
    10 2018-02-05 09:30:00.301            R
    11 2018-02-05 15:59:59.974            R
    12 2018-02-05 16:00:02.438
    13 2018-02-06 09:29:27.279
    14 2018-02-06 09:30:00.045            R
    15 2018-02-06 15:59:59.903            R
    16 2018-02-06 16:01:03.524
    17 2018-02-07 09:29:56.739
    18 2018-02-07 09:30:00.365            H
    19 2018-02-07 15:59:59.940            H
    20 2018-02-07 16:00:00.187
    21 2018-02-08 09:29:28.446
    22 2018-02-08 09:30:00.658            F
    23 2018-02-08 15:59:59.564            F
    24 2018-02-08 16:00:02.355
    25 2018-02-08 19:59:57.061
    """
    if calendar_name is None:
        calendar_name = ''
    if isinstance(calendar_name, str):
        calendar_field_name = ''
    elif isinstance(calendar_name, _Column):
        calendar_field_name = str(calendar_name)
        calendar_name = ''
    else:
        raise ValueError(f"Unsupported type for parameter 'calendar_name': {type(calendar_name)}")
    self.sink(
        otq.MktActivity(calendar_name=calendar_name, calendar_field_name=calendar_field_name)
    )
    self.schema['MKT_ACTIVITY'] = str
    return self


@inplace_operation
def book_diff(self: 'Source', include_initial_book: bool = False, inplace=False) -> Optional['Source']:
    """
    Performs book diffing for every pair of consecutive ticks, each representing a flat book of a fixed depth.

    This method can be thought to be an operation, opposite to :py:meth:`~onetick.py.Source.ob_snapshot_flat`.
    Every input tick, different from the previous one, results in a series of output PRL ticks to be propagated,
    each carrying information about a level deletion, addition or update.

    The input of this method is a time series of flat book ticks,
    just like the result of :py:meth:`~onetick.py.Source.ob_snapshot_flat`.
    More precisely, for a fixed depth N of the flat book,
    input ticks should carry the fields BID_<FIELD_NAME>K and ASK_<FIELD_NAME>K,
    where 1 <= K <= N and <FIELD_NAME> ranges over a specific set of fields F,
    among which PRICE and SIZE are mandatory.

    The output of this method is a time series of PRL ticks,
    carrying information about a level deletion, addition, or update.
    Each tick carries the fields from the above-mentioned set F,
    plus BUY_SELL_FLAG, RECORD_TYPE, TICK_STATUS, and DELETED_TIME.
    BUY_SELL_FLAG carries the side (0 - bid, 1 - ask), the rest are for internal use.

    Parameters
    ----------
    include_initial_book : bool
        This method treats the first tick as an "initial state" of the book.
        If this parameter is set to True, then this initial tick will also be in the output.
    inplace : bool
        The flag controls whether operation should be applied inplace or not.
        If ``inplace=True``, then it returns nothing. Otherwise method returns a new modified
        object.

    Note
    ----
    Tick descriptor changes are not allowed in this event processor.

    Returns
    -------
    :class:`Source` or ``None``

    See Also
    --------
    **BOOK_DIFF** OneTick event processor

    Examples
    --------

    First let's get flat order book snapshot from the database (top level each 12 hours):

    >>> data = otp.DataSource('TRAIN_A_PRL_TRD', tick_type='PRL', symbols='MSFT')  # doctest: +SKIP
    >>> flat = data.ob_snapshot_flat(max_levels=1, bucket_interval=otp.Hour(12))  # doctest: +SKIP
    >>> otp.run(flat, date=otp.dt(2018, 2, 1))  # doctest: +SKIP
                     Time  BID_PRICE1  BID_SIZE1        BID_UPDATE_TIME1  ASK_PRICE1  ASK_SIZE1        ASK_UPDATE_TIME1
    0 2018-02-01 12:00:00       95.53       1400 2018-02-01 11:59:59.797       95.54        200 2018-02-01 11:59:59.978
    1 2018-02-02 00:00:00       94.52        100 2018-02-01 19:57:02.502       94.90        250 2018-02-01 18:35:38.543

    Then we can apply ``book_diff`` method to the result to get the PRL ticks again:

    >>> diff = flat.book_diff()  # doctest: +SKIP
    >>> otp.run(diff, date=otp.dt(2018, 2, 1))  # doctest: +SKIP
            Time  PRICE  SIZE             UPDATE_TIME  BUY_SELL_FLAG RECORD_TYPE  TICK_STATUS         DELETED_TIME
    0 2018-02-02  95.53     0 2018-02-01 11:59:59.797              0           R            0  1969-12-31 19:00:00
    1 2018-02-02  94.52   100 2018-02-01 19:57:02.502              0           R            0  1969-12-31 19:00:00
    2 2018-02-02  94.90   250 2018-02-01 18:35:38.543              1           R            0  1969-12-31 19:00:00
    3 2018-02-02  95.54     0 2018-02-01 11:59:59.978              1           R            0  1969-12-31 19:00:00

    By default the first tick in the query range is not included,
    use parameter ``include_initial_book`` to include it:

    >>> diff = flat.book_diff(include_initial_book=True)  # doctest: +SKIP
    >>> otp.run(diff, date=otp.dt(2018, 2, 1))  # doctest: +SKIP
                     Time  PRICE  SIZE             UPDATE_TIME BUY_SELL_FLAG RECORD_TYPE TICK_STATUS        DELETED_TIME
    0 2018-02-01 12:00:00  95.53  1400 2018-02-01 11:59:59.797             0           R           0 1969-12-31 19:00:00
    1 2018-02-01 12:00:00  95.54   200 2018-02-01 11:59:59.978             1           R           0 1969-12-31 19:00:00
    2 2018-02-02 00:00:00  95.53     0 2018-02-01 11:59:59.797             0           R           0 1969-12-31 19:00:00
    3 2018-02-02 00:00:00  94.52   100 2018-02-01 19:57:02.502             0           R           0 1969-12-31 19:00:00
    4 2018-02-02 00:00:00  94.90   250 2018-02-01 18:35:38.543             1           R           0 1969-12-31 19:00:00
    5 2018-02-02 00:00:00  95.54     0 2018-02-01 11:59:59.978             1           R           0 1969-12-31 19:00:00
    """
    self.sink(otq.BookDiff(include_initial_book=include_initial_book))
    return self


@inplace_operation
def limit(self: 'Source',
          tick_limit: int,
          tick_offset: Optional[int] = None,
          inplace=False) -> Optional['Source']:
    """
    Propagates ticks until the count limit is reached.

    Once the limit is reached,
    hidden ticks will still continue to propagate until the next regular tick appears.

    Parameters
    ----------
    tick_limit: int
        The number of regular ticks to propagate.
        Must be a non-negative integer or -1, which means no limit.
    tick_offset: int
        The number of regular ticks to skip before starting to propagate.
        Must be a non-negative integer.
        By default no ticks are skipped.
    inplace: bool
        The flag controls whether operation should be applied inplace or not.
        If ``inplace=True``, then it returns nothing. Otherwise method returns a new modified
        object.

    Returns
    -------
    :class:`Source` or ``None``

    See Also
    --------
    **LIMIT** OneTick event processor

    Examples
    --------

    Simple example, get first 3 ticks:

    .. testcode::
       :skipif: not otp.compatibility.is_limit_ep_supported()

       data = otp.Ticks(X=[1, 2, 3, 4, 5, 6])
       data = data.limit(3)
       df = otp.run(data)
       print(df)

    .. testoutput::

                            Time  X
       0 2003-12-01 00:00:00.000  1
       1 2003-12-01 00:00:00.001  2
       2 2003-12-01 00:00:00.002  3

    Disable limit by setting it to -1:

    .. testcode::
       :skipif: not otp.compatibility.is_limit_ep_supported()

       data = otp.Ticks(X=[1, 2, 3, 4, 5, 6])
       data = data.limit(-1)
       df = otp.run(data)
       print(df)

    .. testoutput::

                            Time  X
       0 2003-12-01 00:00:00.000  1
       1 2003-12-01 00:00:00.001  2
       2 2003-12-01 00:00:00.002  3
       3 2003-12-01 00:00:00.003  4
       4 2003-12-01 00:00:00.004  5
       5 2003-12-01 00:00:00.005  6

    Setting parameter ``tick_offset`` can be used to skip first ticks before propagating them.

    For example, we can skip first 2 ticks and propagate all other:

    .. testcode::
       :skipif: not otp.compatibility.is_limit_tick_offset_supported()

       data = otp.Ticks(X=[1, 2, 3, 4, 5, 6])
       data = data.limit(-1, tick_offset=2)
       df = otp.run(data)
       print(df)

    .. testoutput::

                            Time  X
       0 2003-12-01 00:00:00.002  3
       1 2003-12-01 00:00:00.003  4
       2 2003-12-01 00:00:00.004  5
       3 2003-12-01 00:00:00.005  6

    Or we can return ticks from the middle of the stream
    by skipping first 2 ticks and then returning next 2 ticks like this:

    .. testcode::
       :skipif: not otp.compatibility.is_limit_tick_offset_supported()

       data = otp.Ticks(X=[1, 2, 3, 4, 5, 6])
       data = data.limit(2, tick_offset=2)
       df = otp.run(data)
       print(df)

    .. testoutput::

                            Time  X
       0 2003-12-01 00:00:00.002  3
       1 2003-12-01 00:00:00.003  4
    """
    if not otp.compatibility.is_limit_ep_supported():
        raise RuntimeError('LIMIT EP isn\'t supported by the current OneTick version.')

    if tick_limit < 0 and tick_limit != -1:
        raise ValueError('Negative values, except -1, not allowed as `tick_limit` in `limit` method.')

    ep_kwargs = {}
    if tick_offset is not None:
        if not isinstance(tick_offset, int) or tick_offset < 0:
            raise ValueError("Parameter 'tick_offset' must be non-negative.")

        if not otp.compatibility.is_limit_tick_offset_supported():
            warnings.warn("Parameter 'tick_offset' is set, but is not supported on this OneTick version")
        else:
            ep_kwargs = {'tick_offset': tick_offset}

    self.sink(otq.Limit(tick_limit=tick_limit, **ep_kwargs))
    return self


def _merge_fields_by_regex(schema: dict, columns_regex: str, match_index: int):
    new_columns: Dict[str, type] = {}
    for column, column_type in list(schema.items()):
        match = re.match(columns_regex, column)
        if match:
            new_column = match.group(match_index)

            if new_column in schema:
                raise KeyError(
                    f'Can\'t apply `show_full_detail` for column `{column}`: '
                    f'column `{new_column}` is already in schema.'
                )

            if new_column in new_columns:
                if column_type != new_columns[new_column]:
                    raise TypeError(
                        f'Can\'t apply `show_full_detail`: '
                        f'type mismatch for columns with suffix `{new_column}`.'
                    )
            else:
                new_columns[new_column] = column_type

            del schema[column]

    schema.update(**new_columns)


@inplace_operation
def virtual_ob(
    self: 'Source',
    quote_source_fields: Optional[List[Union[str, _Column]]] = None,
    quote_timeout: Optional[float] = None,
    show_full_detail: bool = False,
    output_book_format: Literal['ob', 'prl'] = 'ob',
    inplace=False,
) -> Optional['Source']:
    """
    Creates a series of fake orders from a time series of best bids and asks. The algorithm used is as follows:
    For a tick with the best bid or ask, create an order tick adding the new best bid or ask and an order
    withdrawing the old one. Virtual order books can be created for multiple subgroups at once
    by using the ``quote_source_fields`` parameter to specify a list of string fields to be used for grouping.
    A separate book will be created for each combination.

    Parameters
    ----------
    quote_source_fields: Optional[List[Union[str, Column]]]
        Specifies a list of string fields for grouping quotes.
        The virtual order book is then constructed for each subgroup separately and
        the ``SOURCE`` field is constructed to contain the description of the group.
    quote_timeout: Optional[float]
        Specifies the maximum age of a quote that is not stale.
        A quote that is not replaced after more than ``quote_timeout`` seconds is considered stale,
        and delete orders will be generated for it.

        A value of ``quote_timeout`` can be fractional (for example, 3.51)
    show_full_detail: bool
        If set to ``True``, **virtual_ob** will attempt to include
        all fields in the input tick when forming the output tick.

        ``ASK_X/BID_X`` fields will combine under paired field ``X``
        ``ASK_X`` and ``BID_X`` must have the same type.

        If only ``ASK_X`` or ``BID_X`` exist, output will have ``X`` and the missing field
        will be assumed to have its default value.

        Paired and non-paired fields must not interfere with each other and the fields originally added by this EP
    output_book_format: 'prl' or 'ob'
        Supported values are ``prl`` and ``ob``. When set to ``prl``, field ``SIZE`` of output ticks represents
        current size for the tick's source, price, and side, and the EP propagates ``PRICE`` and ``SOURCE``
        as the state keys of its output time series.

        When set to ``ob``, field ``SIZE`` of output ticks represents the delta of size for
        the tick's source, price, and side, and the state key of the output ticks is empty.
    inplace : bool
        The flag controls whether operation should be applied inplace or not.
        If ``inplace=True``, then it returns nothing. Otherwise, method returns a new modified object.

    Returns
    -------
    :class:`Source` or ``None``

    See Also
    --------
    **VIRTUAL_OB** OneTick event processor

    Examples
    --------

    Basic example

    >>> data = otp.DataSource(
    ...     db='US_COMP', symbols='AAPL', tick_type='QTE', date=otp.date(2003, 12, 1)
    ... )  # doctest: +SKIP
    >>> data = data[['ASK_PRICE', 'ASK_SIZE', 'BID_PRICE', 'BID_SIZE']]  # doctest: +SKIP
    >>> data = data.virtual_ob()  # doctest: +SKIP
    >>> otp.run(data)  # doctest: +SKIP
                          Time  PRICE        DELETED_TIME  SIZE  BUY_SELL_FLAG  TICK_STATUS SOURCE
    0  2003-12-01 00:00:00.000  22.28 1969-12-31 19:00:00   500              1            0   AAPL
    1  2003-12-01 00:00:00.000  21.66 1969-12-31 19:00:00   100              0            0   AAPL
    2  2003-12-01 00:00:00.001   21.8 1969-12-31 19:00:00   500              1            0   AAPL
    ...

    Specify columns to group quotes

    >>> data = otp.DataSource(
    ...     db='US_COMP', symbols='AAPL', tick_type='QTE', date=otp.date(2003, 12, 1)
    ... )  # doctest: +SKIP
    >>> data = data[['ASK_PRICE', 'ASK_SIZE', 'BID_PRICE', 'BID_SIZE', 'EXCHANGE']]  # doctest: +SKIP
    >>> data = data.virtual_ob(['EXCHANGE'])  # doctest: +SKIP
    >>> otp.run(data)  # doctest: +SKIP
                          Time  PRICE        DELETED_TIME  SIZE  BUY_SELL_FLAG  TICK_STATUS SOURCE
    0  2003-12-01 00:00:00.000  22.28 1969-12-31 19:00:00   500              1            0      D
    1  2003-12-01 00:00:00.000  21.66 1969-12-31 19:00:00   100              0            0      D
    2  2003-12-01 00:00:00.001   21.8 1969-12-31 19:00:00   500              1            0      P
    ...
    """
    kwargs: Dict[str, Any] = {}

    required_fields = {'BID_PRICE', 'BID_SIZE', 'ASK_PRICE', 'ASK_SIZE'}
    missing_fields = required_fields - set(self.schema.keys())

    if missing_fields:
        raise ValueError('Missing fields required by `virtual_ob`: ' + ', '.join(missing_fields))

    if output_book_format not in {'ob', 'prl'}:
        raise ValueError(
            f'Incorrect value for `output_book_format` parameter: {output_book_format}. '
            f'Supported values: \'ob\' or \'prl\''
        )

    if show_full_detail and output_book_format != 'prl':
        raise ValueError('`output_book_format` should be set to \'prl\' when `show_full_detail` set to `True`')

    is_pnl_and_show_full_detail_supported = is_ob_virtual_prl_and_show_full_detail_supported()

    if show_full_detail and not is_pnl_and_show_full_detail_supported:
        raise RuntimeError('`virtual_ob` not supports `show_full_detail` parameter on this OneTick version')

    if output_book_format != 'ob' and not is_pnl_and_show_full_detail_supported:
        raise RuntimeError('`virtual_ob` not supports `output_book_format` parameter on this OneTick version')

    if is_pnl_and_show_full_detail_supported:
        kwargs['show_full_detail'] = show_full_detail
        kwargs['output_book_format'] = output_book_format

    if quote_source_fields is None:
        quote_source_fields = []

    quote_source_fields_list = list(map(str, quote_source_fields))
    for column in quote_source_fields_list:
        if column not in self.schema:
            raise ValueError(f'Column \'{column}\' passed in `quote_source_fields` parameter is missing in the schema')

    self.sink(
        otq.VirtualOb(
            quote_source_fields=','.join(quote_source_fields_list),
            quote_timeout='' if quote_timeout is None else str(quote_timeout),
            **kwargs,
        )
    )

    schema = {
        'PRICE': float,
        'DELETED_TIME': otp.nsectime,
        'SIZE': float,
        'BUY_SELL_FLAG': int,
        'TICK_STATUS': int,
        'SOURCE': str,
    }

    if show_full_detail:
        exclude_fields = list(required_fields) + quote_source_fields_list
        schema.update({
            k: v for k, v in self.schema.copy().items()
            if k not in exclude_fields
        })

        _merge_fields_by_regex(schema, r'^(ASK|BID)_(.*)$', 2)

    self.schema.set(**schema)

    return self


@copy_signature(otp.functions.corp_actions, add_self=True, drop_parameters=['source'])
def corp_actions(self: 'Source', *args, **kwargs) -> 'Source':
    """
    Adjusts values using corporate actions information loaded into OneTick
    from the reference data file. To use it, location of reference database must
    be specified via OneTick configuration.

    Parameters
    ----------
    adjustment_date : :py:class:`onetick.py.date`, :py:class:`onetick.py.datetime`, int, str, None, optional
        The date as of which the values are adjusted.
        All corporate actions of the types specified in the parameters
        that lie between the tick timestamp and the adjustment date will be applied to each tick.

        This parameter can be either date or datetime .
        `int` and `str` format can be *YYYYMMDD* or *YYYYMMDDhhmmss*.
        When parameter is a date, the time is assumed to be 17:00:00 GMT
        and parameter ``adjustment_date_tz`` is ignored.

        If it is not set, the values are adjusted as of the end date in the query.

        Notice that the ``adjustment date`` is not affected neither by *_SYMBOL_PARAM._PARAM_END_TIME_NANOS*
        nor by the *apply_times_daily* setting in :py:func:`onetick.py.run`.

    adjustment_date_tz : str, optional
        Timezone for ``adjustment date``.

        By default global :py:attr:`tz<onetick.py.configuration.Config.tz>` value is used.
        Local timezone can't be used so in this case parameter is set to GMT.
        When ``adjustment_date`` is in YYYYMMDD format, this parameter is set to GMT.
    fields : str, optional
        A comma-separated list of fields to be adjusted. If this parameter is not set,
        some default adjustments will take place if appropriately named fields exist in the tick:

        - If the ``adjust_rule`` parameter is set to PRICE, and the PRICE field is present,
          it will get adjusted. If the fields ASK_PRICE or BID_PRICE are present, they will get adjusted.
          If fields ASK_VALUE or BID_VALUE are present, they will get adjusted

        - If the ``adjust_rule`` parameter is set to SIZE, and the SIZE field is present,
          it will get adjusted. If the fields ASK_SIZE or BID_SIZE are present, they will get adjusted.
          If fields ASK_VALUE or BID_VALUE are present, they will get adjusted.

    adjust_rule : str, optional
        When set to PRICE, adjustments are applied under the assumption that fields to be adjusted contain prices
        (adjustment direction is determined appropriately).

        When set to SIZE, adjustments are applied under the assumption that fields contain sizes
        (adjustment direction is opposite to that when the parameter's value is PRICE).

        By default the value is PRICE.
    apply_split : bool, optional
        If True, adjustments for splits are applied.
    apply_spinoff : bool, optional
        If True, adjustments for spin-offs are applied.
    apply_cash_dividend : bool, optional
        If True, adjustments for cash dividends are applied.
    apply_stock_dividend : bool, optional
        If True, adjustments for stock dividends are applied.
    apply_security_splice : bool, optional
        If True, adjustments for security splices are applied.
    apply_others : str, optional
        A comma-separated list of names of custom adjustment types to apply.
    apply_all : bool, optional
        If True, applies all types of adjustments, both built-in and custom.

    Returns
    -------
    :py:class:`onetick.py.Source`
        A new source object with applied adjustments.

    See also
    --------
    **CORP_ACTIONS** OneTick event processor

    Examples
    --------
    >>> src = otp.DataSource('US_COMP',
    ...                      tick_type='TRD',
    ...                      start=otp.dt(2022, 5, 20, 9, 30),
    ...                      end=otp.dt(2022, 5, 26, 16))
    >>> df = otp.run(src, symbols='MKD', symbol_date=otp.date(2022, 5, 22))
    >>> df["PRICE"][0]
    0.0911
    >>> src = src.corp_actions(adjustment_date=otp.date(2022, 5, 22),
    ...                        fields="PRICE")
    >>> df = otp.run(src, symbols='MKD', symbol_date=otp.date(2022, 5, 22))
    >>> df["PRICE"][0]
    1.36649931675
    """
    return otp.functions.corp_actions(self, *args, **kwargs)



from pathlib import Path
from setuptools import setup


with open('src/onetick/py/_version.py') as f:
    version = None
    for line in f:
        if line.startswith('#'):
            continue
        _, _, version = line.partition('=')
        version = version.strip().strip("'").strip('"')
    if not version:
        raise RuntimeError("Can't find version in src/onetick/py/_version.py")

with open('requirements.txt') as f:
    required = f.read().splitlines()

strict_requirements = Path('requirements.strict.txt').read_text().splitlines()

setup(name='onetick-py',
      version=version,
      install_requires=required,
      extras_require={
          'strict': strict_requirements,
          'webapi': [
              'onetick.query_webapi',
          ],
          'polars': [
              'polars==1.12.0',
          ]
      },
      entry_points={
          'console_scripts': [
              'onetick-render = onetick.py.utils.render_cli:main',
              'jupyter-onetick_snippets = onetick.doc_utilities.snippets:main',
          ],
      })



from .eps import (
    OqdSourceDprcMain,
    OqdCorpAction,
    OqdCorpDailyFactor,
    OqdSourceBbgbsym,
    OqdSourceBbgbtkr,
    OqdSourceBbgfgc,
    OqdSourceBbgfgs,
    OqdSourceBbgfgv,
    OqdSourceBbgoid,
    OqdSourceCacs,
    OqdSourceCact,
    OqdSourceDes,
    OqdSourceDprcAll,
    OqdSourceDprcExch,
    OqdSourceSho,
    OqdSourceXoid,
    OqdSourceXref,
    OqdSourceXsym,
    OqdTranslate,
)
from .sources import OHLCV, CorporateActions, DescriptiveFields



from onetick.py.otq import otq


class SqlQuery(otq.SqlQuery):
    def __init__(self, sql_statement: str, merge_all_symbols: bool = False, separate_dbname: bool = False):
        """
        Constructs SQL query object.

        Parameters
        ----------
        sql_statement:
            The SQL statement string.
        merge_all_symbols:
            If set to True, ticks returned by the query for all symbols get merged into a single time series.
        separate_dbname:
             If set to True, and ``merge_all_symbols`` is set to True,
             *SYMBOL_NAME* field contains a symbol name without the database name,
             and *DB_NAME* field contains the database name for a symbol.

        See also
        --------
        :py:func:`otp.run <onetick.py.run>`

        Examples
        --------

        Select two fields from a single tick type and symbol and return first three ticks from a single day:

        >>> otp.run(
        ...     otp.SqlQuery("select PRICE,SIZE from US_COMP.TRD"
        ...                  " where symbol_name = 'AAPL'"
        ...                  " and start_time = '2022-03-01 00:00:00 GMT' and end_time = '2022-03-02 00:00:00 GMT'"
        ...                  " limit 3")
        ... )
                             Time  PRICE  SIZE
        0 2022-03-01 00:00:00.000    1.3   100
        1 2022-03-01 00:00:00.001    1.4    10
        2 2022-03-01 00:00:00.002    1.4    50

        Join quotes and trades:

        >>> otp.run(
        ...     otp.SqlQuery("select t.PRICE,q.ASK_PRICE,q.BID_PRICE"
        ...                  " from US_COMP.TRD t join US_COMP.QTE q"
        ...                  " on sametime_as_existing(t.timestamp, q.timestamp, 0) = TRUE"
        ...                  " where t.symbol_name = 'AAPL' and q.symbol_name = 'AAPL'"
        ...                  " and start_time = '2022-03-01 00:00:00 GMT' and end_time = '2022-03-02 00:00:00 GMT'"
        ...                  " limit 2")
        ... )
                             Time  T.PRICE  Q.ASK_PRICE  Q.BID_PRICE
        0 2022-03-01 00:00:00.001      1.4          1.5          1.2
        1 2022-03-01 00:00:00.002      1.4          1.4          1.3

        Calculate average price of trades across several symbols:

        >>> otp.run(
        ...     otp.SqlQuery("select COUNT(*) as COUNT, AVG(PRICE) as AVG_PRICE"
        ...                  " from US_COMP.TRD"
        ...                  " where symbol_name in ('AAPL', 'AAP')"
        ...                  " and start_time = '2022-03-01 00:00:00 GMT' and end_time = '2022-03-02 00:00:00 GMT'",
        ...                  merge_all_symbols=True)
        ... )
                         Time  COUNT  AVG_PRICE
        0 2022-03-01 19:00:00    5.0     18.976
        """
        super().__init__(sql_statement)
        if merge_all_symbols:
            self.set_merge_all_symbols_flag(merge_all_symbols)
        if separate_dbname:
            self.set_separate_dbname_flag(separate_dbname)



# mypy: disable-error-code="attr-defined"

import os
import gc
import inspect
from enum import Enum

if os.getenv('OTP_WEBAPI', default='').lower() not in ('0', 'false', 'no', ''):
    import onetick.query_webapi as otq

    class OneTickLibMock:
        LOGGING_LEVEL_MIN = 0
        LOGGING_LEVEL_LOW = 1
        LOGGING_LEVEL_MEDIUM = 2
        LOGGING_LEVEL_MAX = 3

        def __init__(self, *args, **kwargs):
            pass

        def set_log_file(self, log_file):
            pass

    otq.OneTickLib = OneTickLibMock

else:
    import onetick.query as otq
    import pyomd


class LoggingLevel(Enum):
    MIN = otq.OneTickLib.LOGGING_LEVEL_MIN
    LOW = otq.OneTickLib.LOGGING_LEVEL_LOW
    MEDIUM = otq.OneTickLib.LOGGING_LEVEL_MEDIUM
    MAX = otq.OneTickLib.LOGGING_LEVEL_MAX


class OneTickLib:
    """
    Singleton class for ``otq.OneTickLib`` to initialize it once.

    Returns the same object if it was already initialized.
    """
    __otl_instance = None
    __instance = None
    __args = None

    def __new__(cls, *args, **kwargs):
        if cls.__otl_instance is None:
            cls.__otl_instance = super(OneTickLib, cls).__new__(cls)

            def proxy_wrap(attr, static=False):

                if static:
                    def f(*args, **kwargs):  # type: ignore
                        return getattr(OneTickLib.__instance, attr)(*args, **kwargs)
                    return staticmethod(f)
                else:
                    def f(self, *args, **kwargs):  # type: ignore
                        return getattr(OneTickLib.__instance, attr)(*args, **kwargs)
                    return f

            for attr, value in inspect.getmembers(otq.OneTickLib, callable):
                if not attr.startswith('_') and not hasattr(OneTickLib, attr):
                    fun = inspect.getattr_static(otq.OneTickLib, attr)
                    static = isinstance(fun, staticmethod)
                    setattr(OneTickLib, attr, proxy_wrap(attr, static=static))

        return cls.__otl_instance

    def __init__(self, *args, log_file=None):
        if OneTickLib.__instance is None:
            if not args:
                args = (None,)
            OneTickLib.__args = args
            OneTickLib.__instance = otq.OneTickLib(*args)
            if log_file:
                self.set_log_file(log_file)
        elif args != OneTickLib.__args and args:
            raise ValueError("OneTickLib was already initialized with different "
                             "parameters: Was: {} Now: {}".format(OneTickLib.__args, args))

    def __eq__(self, otl):
        return self.__dict__ == otl.__dict__

    def __ne__(self, otl):
        return self.__dict__ != otl.__dict__

    def __str__(self):
        return "Instance: {}".format(self.__instance.get_one_tick_lib())

    def cleanup(self):
        """
        Destroy otq.OneTickLib instance and reset singleton class
        """
        del OneTickLib.__instance
        gc.collect()
        OneTickLib.__instance = None
        OneTickLib.__args = None

    def set_log_file(self, log_file):
        """
        Set log file for given instance of OneTickLib

        :param log_file: path to log file
        """
        OneTickLib.__instance.set_log_file(str(log_file))
        if hasattr(OneTickLib.__instance, 'close_log_file_in_destructor'):
            # we need to check it to prevent failing CI on the builds that do not have this feature
            OneTickLib.__instance.close_log_file_in_destructor()

    def set_logging_level(self, lvl: LoggingLevel):
        """
        Logging level can be specified by using LoggingLevel Enum class

        :param lvl: available values are LoggingLevel.MIN, LoggingLevel.LOW, LoggingLevel.MEDIUM or LoggingLevel.MAX
        :return:
        """
        OneTickLib.__instance.set_logging_level(lvl)

    def set_authentication_token(self, auth_token: str):
        """
        Set authentication token for given instance of OneTickLib

        :param auth_token: authentication token
        """
        OneTickLib.__instance.set_authentication_token(auth_token)

    @staticmethod
    def override_config_value(config_parameter_name: str, config_parameter_value):
        """
        Override config value of OneTickConfig

        :param config_parameter_name: param to override (could be both set or not set in OneTickConfig)
        :param config_parameter_value: new value of the param
        """
        if OneTickLib.__instance:
            raise RuntimeError('This method should be called before OneTickLib object is constructed to have effect.')
        try:
            pyomd.OneTickLib.override_config_value(config_parameter_name, config_parameter_value)  # noqa
        except NameError:
            raise RuntimeError('This method is not available in WebAPI mode')



from onetick.py.otq import otq

from onetick.py import types as ott
from onetick.py.core.source import Source

from .. import utils

from .common import update_node_tick_type


class RefData(Source):
    def __init__(
        self,
        ref_data_type=None,
        symbol=utils.adaptive,
        db=utils.adaptive_to_default,
        start=utils.adaptive,
        end=utils.adaptive,
        **kwargs,
    ):
        """
        Shows reference data for the specified security and reference data type.

        It can be used to view corporation actions,
        symbol name changes,
        primary exchange info and symbology mapping for a securities,
        as well as the list of symbologies,
        names of custom adjustment types for corporate actions present in a reference database
        as well as names of continuous contracts in database symbology.

        Parameters
        ----------
        ref_data_type: str
            Type of reference data to be queried. Possible values are:

            * corp_actions
            * symbol_name_history
            * primary_exchange
            * symbol_calendar
            * symbol_currency
            * symbology_mapping
            * symbology_list
            * custom_adjustment_type_list
            * all_calendars
            * all_continuous_contract_names

        symbol: str, list of str, :class:`Source`, :class:`query`, :py:func:`eval query <onetick.py.eval>`
            Symbol(s) from which data should be taken.
        db: str
            Name of the database.
        start: :py:class:`otp.datetime <onetick.py.datetime>`
            Start time for tick generation. By default the start time of the query will be used.
        end: :py:class:`otp.datetime <onetick.py.datetime>`
            End time for tick generation. By default the end time of the query will be used.

        See also
        --------
        **REF_DATA** OneTick event processor

        Examples
        --------

        Show calendars for a database TRAIN_A_PRL_TRD in the given range:

        >>> src = otp.RefData(
        ...     'all_calendars',
        ...     db='TRAIN_A_PRL_TRD',
        ...     start=otp.dt(2018, 2, 1),
        ...     end=otp.dt(2018, 2, 9),
        ... )  # doctest: +SKIP
        >>> otp.run(src, symbol_date=otp.dt(2018, 2, 1))  # doctest: +SKIP
                         Time        END_DATETIME CALENDAR_NAME SESSION_NAME SESSION_FLAGS DAY_PATTERN  START_HHMMSS\
          END_HHMMSS TIMEZONE  PRIORITY DESCRIPTION
        0 2018-02-01 00:00:00 2018-02-06 23:59:59          FRED      Regular             R   0.0.12345         93000\
              160000  EST5EDT         0
        1 2018-02-06 23:59:59 2018-02-07 23:59:59          FRED      Holiday             H   0.0.12345         93000\
              160000  EST5EDT         1
        2 2018-02-07 23:59:59 2050-12-31 23:59:59          FRED      Regular             F   0.0.12345         93000\
              160000  EST5EDT         0
        """
        if self._try_default_constructor(**kwargs):
            return

        if ref_data_type is None:
            raise ValueError('Parameter `ref_data_type` was not set')

        if ref_data_type not in [
            'corp_actions', 'symbol_name_history', 'primary_exchange', 'symbol_calendar', 'symbol_currency',
            'symbology_mapping', 'symbology_list', 'custom_adjustment_type_list', 'all_calendars',
            'all_continuous_contract_names',
        ]:
            raise ValueError(f'Incorrect `ref_data_type` value passed: `{ref_data_type}`')

        schema = self._get_schema_for_ref_data_type(ref_data_type)
        ref_data_type = ref_data_type.upper()

        super().__init__(
            _symbols=symbol,
            _start=start,
            _end=end,
            _base_ep_func=lambda: self.base_ep(
                ref_data_type=ref_data_type,
                db=db,
            ),
            schema=schema,
        )

    @staticmethod
    def _get_schema_for_ref_data_type(ref_data_type: str) -> dict:
        if ref_data_type == 'corp_actions':
            return {
                'MULTIPLICATIVE_ADJUSTMENT': float,
                'ADDITIVE_ADJUSTMENT': float,
                'ADJUSTMENT_TYPE': str,
            }
        elif ref_data_type == 'symbol_name_history':
            return {
                'END_DATETIME': ott.nsectime,
                'SYMBOL_NAME': str,
            }
        elif ref_data_type == 'primary_exchange':
            return {
                'END_DATETIME': ott.nsectime,
                'PRIMARY_EXCHANGE': str,
                'SYMBOL_ON_PRIMARY_EXCHANGE': str,
            }
        elif ref_data_type == 'symbol_calendar':
            return {
                'END_DATETIME': ott.nsectime,
                'START_HHMMSS': int,
                'END_HHMMSS': int,
                'SESSION_NAME': str,
                'TIMEZONE': str,
                'DAY_PATTERN': str,
                'SESSION_FLAGS': str,
                'PRIORITY': ott.byte,
                'DESCRIPTION': str,
            }
        elif ref_data_type == 'symbol_currency':
            return {
                'END_DATETIME': ott.nsectime,
                'MULTIPLIER': float,
                'CURRENCY': str,
            }
        elif ref_data_type == 'symbology_mapping':
            return {
                'END_DATETIME': ott.nsectime,
                'MAPPED_SYMBOL_NAME': str,
            }
        elif ref_data_type == 'symbology_list':
            return {
                'SYMBOLOGY_NAME': str,
            }
        elif ref_data_type == 'custom_adjustment_type_list':
            return {
                'ADJUSTMENT_TYPE': str,
            }
        elif ref_data_type == 'all_calendars':
            return {
                'END_DATETIME': ott.nsectime,
                'CALENDAR_NAME': str,
                'SESSION_NAME': str,
                'SESSION_FLAGS': str,
                'DAY_PATTERN': str,
                'START_HHMMSS': int,
                'END_HHMMSS': int,
                'TIMEZONE': str,
                'PRIORITY': ott.byte,
                'DESCRIPTION': str,
            }
        elif ref_data_type == 'all_continuous_contract_names':
            return {
                'CONTINUOUS_CONTRACT_NAME': str,
            }

        raise ValueError('Couldn\'t detect schema: incorrect `ref_data_type` passed')

    def base_ep(
        self,
        ref_data_type,
        db=utils.adaptive_to_default,
    ):
        src = Source(
            otq.RefData(ref_data_type=ref_data_type)
        )

        if db:
            update_node_tick_type(src, tick_type=utils.adaptive, db=db)

        return src



import re

from collections import OrderedDict
from inspect import Parameter
from textwrap import dedent


class Docstring:

    def __init__(self, doc: str):
        if isinstance(doc, list):
            doc = '\n'.join(doc)
        self.doc = doc
        self.indentation = ' ' * (len(doc.lstrip('\n')) - len(doc.lstrip()))
        self.docstring = self._parse()  # NOSONAR

        self.order = ('Description',
                      'Parameters',
                      'Return',
                      'Returns',
                      'Note',
                      'Notes',
                      'Example',
                      'Examples',
                      'See Also')

    def _parse(self):
        def get_prev_name(s: str):
            doc_, name_ = s.rstrip('\n').rsplit('\n', maxsplit=1)
            prefix_ = name_
            name_ = name_.strip()
            prefix_ += '\n' + ' ' * (len(prefix_) - len(name_)) + '-' * len(name_)
            return doc_, name_, prefix_

        docstring = OrderedDict()
        res = re.split(' *?-{4,}', self.doc)  # NOSONAR
        name = 'Description'
        prefix = ''

        # if there is no description part
        if len(res[0].strip().splitlines()) == 1 and len(res) > 1:
            name = res.pop(0).strip()
            prefix = self.indentation + name + '\n' + self.indentation + '-' * len(name)

        for i, part in enumerate(res):
            if i == len(res) - 1:
                docstring[name] = prefix + part
                break
            d, next_name, next_prefix = get_prev_name(part)
            docstring[name] = prefix + d
            name, prefix = next_name, next_prefix
        return docstring

    def get_sig(self, add_self=False):
        """
        Get signature from existing documentation string.
        """
        text = self.docstring.get('Parameters')
        if not text:
            return []
        text = dedent(text)
        text = re.split('-{4,}', text)[1]
        parameters = []
        for line in text.splitlines():
            if not line or line[0] in (' ', '\t'):
                continue
            name, _, _ = line.partition(':')
            if name == 'self' and not add_self:
                continue
            parameters.append(Parameter(name=name, kind=Parameter.POSITIONAL_OR_KEYWORD))
        return parameters

    def __setitem__(self, key, value):
        if key not in self.docstring:
            if key == 'Description':
                self.docstring[key] = ''
            else:
                self.docstring[key] = self.indentation + key + '\n'
                self.docstring[key] += self.indentation + '-' * len(key) + '\n'
        for line in value.split('\n'):
            self.docstring[key] += self.indentation + line + '\n'

    def _apply_order(self):
        key_map = dict((k.lower(), k) for k in self.docstring.keys())
        for k in self.order[::-1]:
            if k.lower() in key_map:
                self.docstring.move_to_end(key_map[k.lower()], last=False)

    def build(self):
        self._apply_order()
        res = ''
        for k, value in self.docstring.items():
            res += value + '\n'

        return res

    def update(self, other: 'Docstring'):
        for key, value in other.docstring.items():

            content = value.split('\n')
            if key == 'Description':
                content = content[1:]
                if not '\n'.join(content).strip():
                    continue
            else:
                content = content[2:]

            if key in self.docstring:
                self.docstring.pop(key)

            for line in content:
                self[key] = line[len(self.indentation):]



from typing import Optional, Tuple

import onetick.py as otp
from onetick.py.otq import otq
import pandas as pd

from onetick.py.core.source import Source

from .. import types as ott
from .. import utils

from .common import update_node_tick_type


class _ReadFromDataFrameSource(Source):
    def __init__(
        self,
        dataframe=None,
        timestamp_column=None,
        symbol_name_field=None,
        symbol_value=None,
        symbol=utils.adaptive,
        db=utils.adaptive_to_default,
        tick_type=utils.adaptive,
        start=utils.adaptive,
        end=utils.adaptive,
        schema=None,
        **kwargs,
    ):
        if self._try_default_constructor(**kwargs):
            return

        if schema is None:
            schema = {}

        super().__init__(
            _symbols=symbol,
            _start=start,
            _end=end,
            _base_ep_func=lambda: self.base_ep(
                dataframe=dataframe,
                timestamp_column=timestamp_column,
                symbol_name_field=symbol_name_field,
                symbol_value=symbol_value,
                db=db,
                tick_type=tick_type,
                columns=schema,
            ),
            schema=schema,
        )

    def base_ep(
        self,
        dataframe,
        timestamp_column=None,
        symbol_name_field=None,
        symbol_value=None,
        db=utils.adaptive_to_default,
        tick_type=utils.adaptive,
        columns=None,
    ):
        if columns is None:
            columns = {}

        if symbol_value is None:
            symbol_value = ''

        if symbol_name_field is None:
            symbol_name_field = ''

        if timestamp_column:
            temp_column_name = f'__TMP_TS_COLUMN__{timestamp_column}__'
            dataframe = dataframe.rename(columns={timestamp_column: temp_column_name})
            timestamp_column = temp_column_name

        src = Source(
            otq.ReadFromDataFrame(
                dataframe=dataframe,
                symbol_name_field=symbol_name_field,
                symbol_value=symbol_value,
            ).get_data_file_ep(),
            schema=columns,
        )

        update_node_tick_type(src, tick_type, db)

        if timestamp_column:
            # DATA_FILE_QUERY process timestamps as GMT timezone
            # In order to process timestamps with query TZ we store timestamps as strings
            # and process them after this EP
            src.sink(
                otq.UpdateField(
                    field="TIMESTAMP",
                    value=f'parse_nsectime("%Y-%m-%d %H:%M:%S.%J", {timestamp_column}, _TIMEZONE)',
                    allow_unordered_output_times=True,
                )
            )
            src.sink(otq.Passthrough(fields=timestamp_column, drop_fields=True))

        return src


def _get_offsets(dataframe, timestamp_column) -> Tuple[otp.datetime, list]:
    offsets = [ott.timedelta(0)]
    base_ts = pd.Timestamp(dataframe[timestamp_column][0])

    for i in range(1, len(dataframe)):
        diff: pd.Timedelta = pd.Timestamp(dataframe[timestamp_column][i]) - base_ts
        offsets.append(ott.timedelta(**diff.components._asdict()))

    return ott.datetime(base_ts), offsets


def _fallback_source(
    dataframe,
    timestamp_column=None,
    symbol_value=None,
    symbol_name_field=None,
    db=utils.adaptive_to_default,
    tick_type=utils.adaptive,
):
    rows_num = len(dataframe)
    if symbol_value:
        dataframe['SYMBOL_NAME'] = symbol_value

    if symbol_name_field:
        dataframe['SYMBOL_NAME'] = dataframe[symbol_name_field]

    data = dataframe.to_dict(orient='list')

    ticks_kwargs = {}
    if timestamp_column:
        if rows_num > 0:
            base_ts, offsets = _get_offsets(dataframe, timestamp_column)
            ticks_kwargs['offset'] = offsets
            ticks_kwargs['start'] = base_ts

        # remove original timestamp
        del data[timestamp_column]

    ts_column_mapping = []
    save_ts_column_list = [
        col for col in data.keys()
        if col and col.lower() == 'timestamp'
    ]
    if save_ts_column_list:
        # For some reason OneTick CSV_FILE_LISTING doesn't like timestamp columns with any case

        for idx, col in enumerate(save_ts_column_list):
            _mapping = (f'__TMP_TIMESTAMP_COLUMN_{idx}__', col)
            ts_column_mapping.append(_mapping)
            data[_mapping[0]] = data[_mapping[1]]
            del data[_mapping[1]]

    src = otp.Ticks(data=data, db=db, tick_type=tick_type, **ticks_kwargs)

    if symbol_name_field:
        src.drop(['SYMBOL_NAME'], inplace=True)

    if not timestamp_column:
        src['Time'] = otp.meta_fields.end_time

    if ts_column_mapping:
        for _mapping in ts_column_mapping:
            src.rename({_mapping[0]: _mapping[1]}, inplace=True)

    return src


def _autodetect_timestamp_column(dataframe: pd.DataFrame) -> Optional[str]:
    timestamp_columns = [col for col in dataframe.columns if col.lower() in ['time', 'timestamp']]
    if len(timestamp_columns) > 1:
        raise ValueError(
            'Could not determine timestamp column from multiple available choices: ' + ', '.join(timestamp_columns)
        )
    elif len(timestamp_columns) == 1:
        return timestamp_columns[0]

    return None


def ReadFromDataFrame(
    dataframe=None,
    timestamp_column=utils.adaptive,
    symbol_name_field=None,
    symbol=utils.adaptive,
    db=utils.adaptive_to_default,
    tick_type=utils.adaptive,
    start=utils.adaptive,
    end=utils.adaptive,
    force_compatibility_mode=False,
    **kwargs,
):
    """
    Load :pandas:`pandas.DataFrame` as data source

    Parameters
    ----------
    dataframe: :pandas:`pandas.DataFrame`
        Pandas DataFrame to load.
    timestamp_column: str, optional
        Column containing time info.

        If parameter not set and DataFrame has one of columns ``TIME`` or ``Timestamp`` (case-insensitive),
        it will be automatically used as ``timestamp_column``. To disable this, set ``timestamp_column=None``.

        Timestamp column dtype should be either datetime related or string.
    symbol_name_field: str, optional
        Column containing symbol name.
    symbol: str
        Symbol(s) from which data should be taken.

        If both `symbol_name_field` and `symbol` are omitted
        :py:attr:`otp.config.default_symbol<onetick.py.configuration.Config.default_symbol>` value will be used.
    db: str
        Custom database name for the node of the graph.
    tick_type: str
        Tick type.
        Default: ANY.
    start: :py:class:`otp.datetime <onetick.py.datetime>`
        Custom start time of the query.
    end: :py:class:`otp.datetime <onetick.py.datetime>`
        Custom end time of the query.
    force_compatibility_mode: bool
        Force use of old dataframe load method

    Examples
    --------

    Let's assume that we have the following pandas dataframe:

    >>> print(dataframe)  # doctest: +SKIP
                     Timestamp  SIDE  PRICE  SIZE
    0  2024-01-01 12:00:00.001   BUY  50.05   100
    1  2024-01-01 12:00:02.000  SELL  50.05   150
    2  2024-01-01 12:00:02.500   BUY  49.95   200
    3  2024-01-01 12:00:03.100  SELL  49.98    80
    4  2024-01-01 12:00:03.250   BUY  50.02   250

    Simple dataframe loading, timestamp column will be automatically detected and converted to datetime:

    >>> src = otp.ReadFromDataFrame(dataframe, symbol='AAPL')  # doctest: +SKIP
    >>> otp.run(src, date=otp.date(2024, 1, 1))  # doctest: +SKIP
                         Time  SIDE  PRICE  SIZE SYMBOL_NAME
    0 2024-01-01 12:00:00.001   BUY  50.05   100        AAPL
    1 2024-01-01 12:00:02.000  SELL  50.05   150        AAPL
    2 2024-01-01 12:00:02.500   BUY  49.95   200        AAPL
    3 2024-01-01 12:00:03.100  SELL  49.98    80        AAPL
    4 2024-01-01 12:00:03.250   BUY  50.02   250        AAPL

    Setting custom `timestamp_column`. For example, if we have ``DATA_TIMES`` column, instead of ``Timestamps``

    >>> src = otp.ReadFromDataFrame(dataframe, symbol='AAPL', timestamp_column='DATA_TIMES')  # doctest: +SKIP
    >>> otp.run(src, date=otp.date(2024, 1, 1))  # doctest: +SKIP
                         Time  SIDE  PRICE  SIZE SYMBOL_NAME
    0 2024-01-01 12:00:00.001   BUY  50.05   100        AAPL
    1 2024-01-01 12:00:02.000  SELL  50.05   150        AAPL
    2 2024-01-01 12:00:02.500   BUY  49.95   200        AAPL
    3 2024-01-01 12:00:03.100  SELL  49.98    80        AAPL
    4 2024-01-01 12:00:03.250   BUY  50.02   250        AAPL

    You can load data even without time data. ``Time`` column will be set as query end time.

    >>> src = otp.ReadFromDataFrame(dataframe, symbol='AAPL')  # doctest: +SKIP
    >>> otp.run(src, date=otp.date(2024, 1, 1))  # doctest: +SKIP
            Time  SIDE  PRICE  SIZE SYMBOL_NAME
    0 2024-01-02   BUY  50.05   100        AAPL
    1 2024-01-02  SELL  50.05   150        AAPL
    2 2024-01-02   BUY  49.95   200        AAPL
    3 2024-01-02  SELL  49.98    80        AAPL
    4 2024-01-02   BUY  50.02   250        AAPL

    Same effect will be if you don't set ``timestamp_column`` and disable automatic timestamp detection:

    >>> src = otp.ReadFromDataFrame(dataframe, symbol='AAPL', timestamp_column=None)  # doctest: +SKIP
    >>> otp.run(src, date=otp.date(2024, 1, 1))  # doctest: +SKIP
            Time                Timestamp  SIDE  PRICE  SIZE SYMBOL_NAME
    0 2024-01-02  2024-01-01 12:00:00.001   BUY  50.05   100        AAPL
    1 2024-01-02  2024-01-01 12:00:02.000  SELL  50.05   150        AAPL
    2 2024-01-02  2024-01-01 12:00:02.500   BUY  49.95   200        AAPL
    3 2024-01-02  2024-01-01 12:00:03.100  SELL  49.98    80        AAPL
    4 2024-01-02  2024-01-01 12:00:03.250   BUY  50.02   250        AAPL

    Setting ``symbol_name_field`` for setting symbol name from dataframe.
    In this example, let's say, that we have column ``SYMBOL`` with symbol names.

    >>> src = otp.ReadFromDataFrame(dataframe, symbol_name_field='SYMBOL')  # doctest: +SKIP
    >>> otp.run(src, date=otp.date(2024, 1, 1))  # doctest: +SKIP
                         Time  SIDE  PRICE  SIZE SYMBOL
    0 2024-01-01 12:00:00.001   BUY  50.05   100   AAPL
    1 2024-01-01 12:00:02.000  SELL  50.05   150   AAPL
    2 2024-01-01 12:00:02.500   BUY  49.95   200   AAPL
    3 2024-01-01 12:00:03.100  SELL  49.98    80   AAPL
    4 2024-01-01 12:00:03.250   BUY  50.02   250   AAPL
    """
    if dataframe is None:
        raise ValueError('DataFrame should be passed to `ReadFromDataFrame` constructor')

    if not isinstance(dataframe, pd.DataFrame):
        raise ValueError(f'`dataframe` parameter expected to be pandas DataFrame, got `{type(dataframe)}`')

    dataframe = dataframe.copy(deep=True)

    if ('TIMESTAMP' in dataframe.columns or 'Time' in dataframe.columns) and not (
        timestamp_column in ['TIMESTAMP', 'Time'] or timestamp_column is utils.adaptive
    ):
        # Can't set meta fields
        raise ValueError(
            'It\'s not allowed to both have `TIMESTAMP` or `Time` column in DataFrame '
            'and pass `timestamp_column` parameter with different column '
            'or disable timestamp column autodetection'
        )

    if timestamp_column is utils.adaptive:
        timestamp_column = _autodetect_timestamp_column(dataframe)

    if timestamp_column:
        if timestamp_column not in dataframe.columns:
            raise ValueError(f'Column `{timestamp_column}` passed as `timestamp_column` parameter not in dataframe')

        if ott.np2type(dataframe[timestamp_column].dtype) in [ott.nsectime, ott.msectime]:
            # convert back to string
            dataframe[timestamp_column] = (dataframe[timestamp_column].dt.strftime('%Y-%m-%d %H:%M:%S.%f') +
                                           dataframe[timestamp_column].dt.nanosecond.astype(str).str.zfill(3))

    columns = {}
    for column, dtype in dataframe.dtypes.to_dict().items():
        if timestamp_column == column:
            continue
        else:
            dtype = ott.np2type(dtype)

        columns[column] = dtype

    if not symbol_name_field and symbol is utils.adaptive:
        symbol = otp.config.get('default_symbol')

    symbol_value = None
    if symbol is not utils.adaptive:
        if symbol_name_field:
            raise ValueError('`symbol_name_field` parameter is passed while `symbol` parameter is defined')

        symbol_value = symbol

        # otq.ReadFromDataFrame adds this column in this case
        columns['SYMBOL_NAME'] = str
    elif symbol_name_field and symbol_name_field not in dataframe.columns:
        raise ValueError(f'Column `{symbol_name_field}` passed as `symbol_name_field` parameter not in dataframe')

    if hasattr(otq, 'ReadFromDataFrame') and not force_compatibility_mode:
        return _ReadFromDataFrameSource(
            dataframe=dataframe,
            timestamp_column=timestamp_column,
            symbol_name_field=symbol_name_field,
            symbol_value=symbol_value,
            symbol=symbol,
            db=db,
            tick_type=tick_type,
            start=start,
            end=end,
            schema=columns,
            **kwargs,
        )
    else:
        return _fallback_source(
            dataframe=dataframe,
            timestamp_column=timestamp_column,
            symbol_name_field=symbol_name_field,
            symbol_value=symbol_value,
        )



from typing import TYPE_CHECKING, Any, List, Optional

from onetick.py.otq import otq

from .misc import inplace_operation

if TYPE_CHECKING:
    from onetick.py.core.source import Source


@inplace_operation
def drop(self: 'Source', columns: List[Any], inplace=False) -> Optional['Source']:
    r"""
    Remove a list of columns specified by names or regular expressions from the Source.

    If a column with such name wasn't found the error will be raised,
    if the regex was specified and there aren't any matched columns, do nothing.

    Regex is any string containing any of characters ***+?\:[]{}()**,
    dot is a valid symbol for OneTick identifier,
    so **a.b** will be passed to OneTick as an identifier,
    if you want specify such regex use parenthesis - **(a.b)**

    If a string without special characters or :py:class:`~onetick.py.Column` object is specified in the list,
    then it is assumed that this is a literal name of the column, not regular expression.

    Parameters
    ----------
    columns : str, Column or list of them
        Column(s) to remove. You could specify a regex or collection of regexes, in such case columns with
        match names will be deleted.
    inplace: bool
        A flag controls whether operation should be applied inplace.
        If inplace=True, then it returns nothing. Otherwise method
        returns a new modified object.

    Returns
    ----------
    :class:`Source` or ``None``

    See Also
    --------
    **PASSTHROUGH** OneTick event processor

    Examples
    --------

    Drop a single field:

    >>> data = otp.Ticks(X1=[1, 2, 3],
    ...                  X2=[3, 2, 1],
    ...                  A1=["A", "A", "A"])
    >>> data = data.drop("X1")  # OTdirective: snippet-name: Arrange.drop.one field;
    >>> otp.run(data)
                         Time  X2 A1
    0 2003-12-01 00:00:00.000   3  A
    1 2003-12-01 00:00:00.001   2  A
    2 2003-12-01 00:00:00.002   1  A

    Regexes also could be specified in such case all matched columns will be deleted:

    >>> data = otp.Ticks(X1=[1, 2, 3],
    ...                  X2=[3, 2, 1],
    ...                  A1=["A", "A", "A"])
    >>> data = data.drop(r"X\d+")   # OTdirective: snippet-name: Arrange.drop.regex;
    >>> otp.run(data)
                         Time A1
    0 2003-12-01 00:00:00.000  A
    1 2003-12-01 00:00:00.001  A
    2 2003-12-01 00:00:00.002  A

    Both column names and regular expressions can be specified at the same time:

    >>> # OTdirective: snippet-name: Arrange.drop.multiple;
    >>> data = otp.Ticks(X1=[1, 2, 3],
    ...                  X2=[3, 2, 1],
    ...                  Y1=[1, 2, 3],
    ...                  Y2=[3, 2, 1],
    ...                  Y11=["a", "b", "c"],
    ...                  Y22=["A", "A", "A"])
    >>> data = data.drop([r"X\d+", "Y1", data["Y2"]])
    >>> otp.run(data)
                         Time Y11 Y22
    0 2003-12-01 00:00:00.000   a   A
    1 2003-12-01 00:00:00.001   b   A
    2 2003-12-01 00:00:00.002   c   A

    **a.b** will be passed to OneTick as an identifier,
    if you want specify such regex use parenthesis - **(a.b)**:

    >>> data = otp.Ticks({"COLUMN.A": [1, 2, 3], "COLUMN1A": [3, 2, 1],
    ...                   "COLUMN1B": ["a", "b", "c"], "COLUMN2A": ["c", "b", "a"]})
    >>> data = data.drop("COLUMN.A")    # OTdirective: skip-snippet:;
    >>> otp.run(data)
                         Time  COLUMN1A COLUMN1B COLUMN2A
    0 2003-12-01 00:00:00.000         3        a        c
    1 2003-12-01 00:00:00.001         2        b        b
    2 2003-12-01 00:00:00.002         1        c        a

    >>> data = otp.Ticks({"COLUMN.A": [1, 2, 3], "COLUMN1A": [3, 2, 1],
    ...                   "COLUMN1B": ["a", "b", "c"], "COLUMN2A": ["c", "b", "a"]})
    >>> data = data.drop("(COLUMN.A)")  # OTdirective: skip-snippet:;
    >>> otp.run(data)
                         Time COLUMN1B
    0 2003-12-01 00:00:00.000        a
    1 2003-12-01 00:00:00.001        b
    2 2003-12-01 00:00:00.002        c
    """
    self.__delitem__(columns)
    return self


def __delitem__(self: 'Source', obj):
    if isinstance(obj, (list, tuple)):
        objs = obj
    else:
        objs = (obj,)

    items_to_passthrough, regex = self._columns_names_regex(objs, drop=True)
    self.sink(otq.Passthrough(drop_fields=True, fields=",".join(items_to_passthrough), use_regex=regex))



from datetime import datetime


DEFAULT_START_DATE = datetime(year=2003, month=1, day=1)
DEFAULT_END_DATE = datetime(year=2099, month=12, day=31)


class compression_type:
    LZ4 = "LZ4"
    GZIP = "GZIP"
    NATIVE = "NATIVE"
    NATIVE_PLUS_LZ4 = "NATIVE_PLUS_LZ4"
    NATIVE_PLUS_GZIP = "NATIVE_PLUS_GZIP"
    NATIVE_PLUS_ZSTD = "NATIVE_PLUS_ZSTD"


class access_method:
    FILE = "file"
    MEMORY = "memory"
    SOCKET = "socket"



import operator
import os

import onetick.py as otp
from onetick.py.otq import otq

from onetick.py.core.source import Source

from .. import types as ott
from .. import utils, configuration
from ..core import query_inspector
from ..core.column_operations.base import _Operation


_QUERY_PARAM_SPECIAL_CHARACTERS = "=,"


class Query(Source):
    def __init__(
        self,
        query_object=None,
        out_pin=utils.adaptive,
        symbol=utils.adaptive,
        start=utils.adaptive,
        end=utils.adaptive,
        params=None,
        schema=None,
        **kwargs,
    ):
        """
        Create data source object from .otq file or query object

        Parameters
        ----------
        query_object: path or :class:`query`
            query to use as a data source
        out_pin: str
             query output pin name
        symbol: None, :py:class:`onetick.py.adaptive`
            Symbol(s) from which data should be taken.
        start, end : :py:class:`datetime.datetime`, :py:class:`otp.datetime <onetick.py.datetime>` or utils.adaptive
            Time interval from which the data should be taken.
        params: dict
            params to pass to query.
            Only applicable to string ``query_object``
        """
        if self._try_default_constructor(schema=schema, **kwargs):
            return

        if params is None:
            params = {}

        # Ignore because of the "Only @runtime_checkable protocols can be used with instance and class checks"
        if isinstance(query_object, (str, os.PathLike)):  # type: ignore
            query_object = query(str(query_object), **params)
        elif isinstance(query_object, query):
            if len(params) > 0:
                raise ValueError("Cannot pass both params and a query() (not str) query_object parameter")
        else:
            raise ValueError("query_object parameter has to be either a str (path to the query) or a query object")

        if symbol == utils.adaptive:
            if query_object.graph_info is None or not query_object.graph_info.has_unbound_sources:
                symbol = None
        elif symbol is not None:
            raise ValueError("symbol parameter should be either None or otp.adaptive")

        super().__init__(
            _symbols=symbol, _start=start, _end=end, _base_ep_func=lambda: self.base_ep(query_object, out_pin),
            schema=schema, **kwargs,
        )

    def base_ep(self, query_object, out_pin):
        nested = otq.NestedOtq(query_object.path, query_object.str_params)
        graph = query_object.graph_info

        if out_pin is utils.adaptive:
            if graph is not None:
                if len(graph.nested_outputs) == 1:
                    return Source(nested[graph.nested_outputs[0].NESTED_OUTPUT])
                if len(graph.nested_outputs) > 1:
                    raise ValueError(
                        f'Query "{query_object.query_name}" has multiple outputs, but you have not '
                        "specified which one should be used. You could specify it"
                        ' using "out_pin" parameter of the Query constructor.'
                    )
            # no output
            return Source(nested, _has_output=False)

        if graph is not None:
            existed_out_pins = set(map(operator.attrgetter("NESTED_OUTPUT"), graph.nested_outputs))
            if out_pin not in existed_out_pins:
                raise ValueError(
                    f'Query "{query_object.query_name}" does not have the "{out_pin}" output, there are only following '
                    f"output pins exist: {','.join(existed_out_pins)}"
                )
        return Source(nested[out_pin])


class query:
    """
    Constructs a query object with a certain path.
    Keyword arguments specify query parameters.

    You also can pass an instance of ``otp.query.config`` class as the second positional argument to
    specify a query.

    Parameters
    ----------

    path : str
        path to an .otq file.
        If path is relative, then it's assumed that file is located in one of the directories
        specified in OneTick ``OTQ_FILE_PATH`` configuration variable.
        If there are more than one query in the file, then its name should be specified
        in the format ``<path>::<query-name>``.

        Also prefix ``remote://<database-name>::`` can be used to specify if query is located
        on the remote server.
        If such path exists locally too, then this file is inspected to get info about query and its pins
        as we can't get this info from remote server.
    config:
        optional ``otp.query.config`` object.
        This object can be used to specify different query options, e.g. output columns.
    params:
        parameters for the query.
        Dictionary if parameters' names and their values.

    Raises
    ------
    ValueError, TypeError

    Examples
    --------

    Adding local query and applying it to the source:

    >>> q = otp.query('/otqs/some.otq::some_query', PARAM1='val1', PARAM2=3.14)  # doctest: +SKIP
    >>> t = otp.Tick(A=1)
    >>> t = t.apply(q)  # doctest: +SKIP

    Adding remote query:

    >>> otp.query('remote://DATABASE::/otqs/some.otq::some_query', PARAM1='val1', PARAM2=3.14)  # doctest: +ELLIPSIS
    <onetick.py.sources.query.query object at ...>

    Creating python wrapper function around query from ``.otq`` file:

    >>> def add_w(src):
    ...     schema = dict(src.schema)
    ...     if 'W' in schema:
    ...         raise ValueError("column 'W' already exists")
    ...     else:
    ...         schema['W'] = str
    ...     q = otp.query('add_w.otq::w', otp.query.config(output_columns=list(schema.items())))
    ...     return src.apply(q)
    >>> t = otp.Tick(A=1)
    >>> t = add_w(t)
    >>> t.schema
    {'A': <class 'int'>, 'W': <class 'str'>}
    >>> t['X'] = t['W'].str.upper()
    >>> otp.run(t)
            Time  A      W      X
    0 2003-12-01  1  hello  HELLO
    """

    class config:
        """
        The config allows to specify different query options.
        """

        _special_values = {"input"}

        def __init__(self, output_columns=None):
            """
            Parameters
            ----------

            output_columns : str, list, dict, optional
                The parameter defines what the outputs columns are.
                Default value is ``None`` that means no output fields after applying query
                for every output pin.

                The ``input`` string value means that output columns are the same as inputs for
                every output pin.

                A list of tuples allows to define output columns with their types;
                for example ``[('x', int), ('y', float), ...]``. Applicable for every output
                pin.

                A dict allows to specify output columns for every output pin.

            Raises
            ------
            TypeError, ValueError

            Examples
            --------
            >>> otp.query('/otqs/some.otq::some_query', otp.query.config(output_columns=[('X': int)]))  # doctest: +SKIP
            """

            if output_columns is not None:
                if isinstance(output_columns, list):
                    self._validate_columns(output_columns)
                elif isinstance(output_columns, dict):
                    for pin, columns in output_columns.items():
                        if not isinstance(pin, str):
                            raise TypeError(f"Name of pin '{type(pin)}' is of non-str type '%s'")
                        else:
                            self._validate_columns(columns)

                elif not isinstance(output_columns, str):
                    raise TypeError(f'"output_columns" does not support value of the "{type(output_columns)}" type')

                if isinstance(output_columns, str):
                    if output_columns not in self._special_values:
                        raise ValueError(f'Config does not support "{output_columns}" value')

            self.output_columns = output_columns

        def _validate_list_item(self, item):
            if isinstance(item, str):
                if item not in self._special_values:
                    raise ValueError(f"Value {item} is not supported.")

            else:
                if not isinstance(item, (tuple, list)) or (len(item) != 2) or not isinstance(item[0], str):
                    raise TypeError("Value %s is not a name-type tuple.")

        def _validate_columns(self, columns):
            if isinstance(columns, str):
                if columns not in self._special_values:
                    raise ValueError(f"A pin has invalid output columns definition: '{columns}'")

            elif isinstance(columns, list):
                if columns.count("input") > 1:
                    raise ValueError(f"More than one 'input' value in {columns}")

                for item in columns:
                    self._validate_list_item(item)

            else:
                raise TypeError(f"A pin's columns definition is of unsupported type '{type(columns)}'")

        def _get_output_columns_for_pin(self, out_pin_name):
            if isinstance(self.output_columns, dict):
                if out_pin_name not in self.output_columns:
                    raise ValueError(f"Pin {out_pin_name} wasn't declared in the config")
                else:
                    return self.output_columns[out_pin_name]

            else:
                return self.output_columns

        def _apply(self, out_pin_name, src):
            """
            Applying specified logic on a certain object. Used internally in the functions.apply_query
            """
            columns_descriptor = self._get_output_columns_for_pin(out_pin_name)
            if columns_descriptor is None:
                # drop columns by default, because we don't know
                # how an external query changes data schema
                src.drop_columns()
            elif columns_descriptor != "input":
                if "input" not in columns_descriptor:
                    src.drop_columns()

                for item in columns_descriptor:
                    if item != "input":
                        name, dtype = item
                        src.schema[name] = dtype

    def __init__(self, path, *config, **params):

        # prepare parameters
        self._str_params = None
        self.params = params
        self.update_params()

        # prepare configs
        if len(config) > 1:
            raise ValueError(f"It is allowed to specify only one config object, but passed {len(config)}")
        elif len(config) == 1:
            if not isinstance(config[0], self.config):
                raise TypeError(
                    f'It is expected to see config of the "query.config" type, but got "{type(config[0])}"'
                )
            self.config = config[0]
        else:
            self.config = self.config()

        # prepare path and query name

        path = str(path)

        remote = path.startswith('remote://')
        if remote:
            self.path = path
            _, path = path.split('::', maxsplit=1)
        else:
            if otp.__webapi__:
                # remote:// used only for remote queries, not for locals, like we do without webapi
                self.path = path
            else:
                self.path = f"remote://{configuration.config.get('default_db', 'LOCAL')}::" + path

        self.query_path, self.query_name = utils.query_to_path_and_name(path)

        # if query_path does not exist, then we try
        # to resolve it with OTQ_FILE_PATH assuming that
        # a relative path is passed
        if not os.path.exists(self.query_path):
            otq_path = utils.get_config_param(os.environ["ONE_TICK_CONFIG"], "OTQ_FILE_PATH", "")
            try:
                self.query_path = utils.abspath_to_query_by_otq_path(otq_path, self.query_path)
            except FileNotFoundError:
                if remote:
                    # TODO: we want to get self.graph_info from remote query somehow, probably will have to download it
                    self.graph_info = None
                    return
                raise

        if self.query_name is None:
            # it seems that query name was not passed, then try to find it
            queries = query_inspector.get_queries(self.query_path)
            if len(queries) > 1:
                raise ValueError(f"{self.query_path} has more than one query, "
                                 f"but you have not specified which one to use.")
            self.query_name = queries[0]

        self.graph_info = query_inspector.get_query_info(self.query_path, self.query_name)

    def __call__(self, *ticks, **pins):
        """
        Return object representing outputs of the query.
        This object can be used to get a specified output pin of the query as a new :py:class:`onetick.py.Source`.

        Examples
        --------
        >>> query = otp.query('/otqs/some.otq::some_query', PARAM1='val1')  # doctest: +SKIP
        >>> query()['OUT']  # doctest: +SKIP,+ELLIPSIS
        <onetick.py.core.source.Source at ...>
        """
        for key, value in pins.items():
            if not isinstance(value, Source):
                raise ValueError(f'Input "{key}" pin does not support "{type(value)}" type')

        if self.graph_info is not None and len(pins) == 0 and len(ticks) == 1:
            if len(self.graph_info.nested_inputs) != 1:
                raise ValueError(
                    f'It is expected the query "{self.query_path}" to have one input, but it'
                    f" has {len(self.graph_info.nested_inputs)}"
                )

            pins[self.graph_info.nested_inputs[0].NESTED_INPUT] = ticks[0]
        elif len(ticks) == 0:
            # it is the valid case, when query has no input pins
            pass
        else:
            raise ValueError("It is allowed to pass only one non-specified input")

        outputs = self._outputs()
        outputs.query = self
        outputs.in_sources = pins

        return outputs

    class _outputs:
        def __getitem__(self, key):
            output_pins = []

            if isinstance(key, tuple):
                output_pins = list(key)
            elif isinstance(key, str):
                output_pins = [key]
            elif key is None:
                # No output
                pass
            else:
                raise ValueError(f'Output pins can not be of "{type(key)}" type')

            return otp.apply_query(
                self.query, in_sources=self.in_sources, output_pins=output_pins, **self.query.params
            )

    def to_eval_string(self):
        """
        Converts query object to OneTick's `eval` string.
        """
        res = '"' + self.path + '"'
        if self.params:
            res += f', "{self._params_to_str(self.params, with_expr=True)}"'
        return "eval(" + res + ")"

    def update_params(self, **new_params):
        """
        Update dictionary of parameters of the query.
        """
        if new_params:
            self.params.update(new_params)

    @property
    def str_params(self):
        """
        Query parameters converted to OneTick string representation.
        """
        if self._str_params is None:
            self._str_params = self._params_to_str(self.params)
        return self._str_params

    @staticmethod
    def _params_to_str(params, *, with_expr=False):
        """ converts param to str

        Parameters
        ----------
        params: dict
            Parameters as dict(name=value)
        with_expr:
            If true return all expression in expr() function

        Returns
        -------
        result: str
            string representation of parameters ready for query evaluation
        """

        def to_str(v):
            if isinstance(v, list):
                return "\\,".join(map(to_str, v))
            else:
                if with_expr:
                    is_dt = ott.is_time_type(v)
                    if is_dt:
                        v = ott.value2str(v)
                    result = query._escape_quotes_in_eval(v)
                    if isinstance(v, _Operation) and getattr(v, "name", None) != "_SYMBOL_NAME" or is_dt:
                        result = f"expr({result})"
                else:
                    result = query._escape_characters_in_query_param(str(v))
                return result

        return ",".join(key + "=" + to_str(value) for key, value in params.items())

    @staticmethod
    def _escape_quotes_in_eval(v):
        return str(v).translate(str.maketrans({"'": r"\'", '"': r'\"'}))

    @staticmethod
    def _escape_characters_in_query_param(result):
        # 0 - no need to add backslash, 1 - need to add
        char_map = [0] * len(result)

        # put 1 between two quotes symbols
        open_char = None
        last_inx = 0
        for inx, c in enumerate(result):
            if open_char == c:
                open_char = None
                continue

            if not open_char and c in ("'", '"'):
                open_char = c
                last_inx = inx + 1
                continue

            if open_char:
                char_map[inx] = 1

        # clean open tail if necessary
        if open_char:
            char_map[last_inx:] = [0] * (len(result) - last_inx)

        # apply mapping
        res = []
        last_esc = False  # do not add esc if the previous one is already esc
        n_brackets_in_expr_block = 0  # do not escape in expr(...)
        for inx, c in enumerate(result):
            if c == "(":
                if n_brackets_in_expr_block:
                    n_brackets_in_expr_block += 1
                elif result[inx - 4:inx] == "expr":
                    n_brackets_in_expr_block = 1
            if c == ")" and n_brackets_in_expr_block:
                n_brackets_in_expr_block -= 1

            if c in _QUERY_PARAM_SPECIAL_CHARACTERS and char_map[inx] == 0:
                if not last_esc and not n_brackets_in_expr_block:
                    c = "\\" + c

            last_esc = c == "\\"

            res.append(c)

        return "".join(res)



import inspect
import string

from .column_operations.base import _Operation
from .. import types as ott


_allowed_field_name_chars = set(string.ascii_letters) | set(string.digits) | {'_', '.'}


def validate_onetick_field_name(field_name):
    if len(field_name) < 1 or len(field_name) > 127:
        return False
    chars = set(char for char in field_name)

    return chars.issubset(_allowed_field_name_chars)


def field_name_contains_lowercase(field_name):
    for char in field_name:
        if char.islower():
            return True
    return False


class Column(_Operation):
    """
    :py:class:`~onetick.py.Source` column container.

    This is the object you get when using :py:meth:`~onetick.py.Source.__getitem__`.
    You can use this object everywhere where :py:class:`~onetick.py.Operation` object can be used.

    Examples
    --------
    >>> t = otp.Tick(A=1)
    >>> t['A']
    Column(A, <class 'int'>)
    """

    @staticmethod
    def _check_name(name):
        if not validate_onetick_field_name(name):
            raise ValueError(f"Field name '{name}' is not a valid field name. "
                             "Onetick field names can contain upper and lower English letters, digits, "
                             "symbols '.' and '_'. Also, Onetick field names cannot be longer than 127 characters.")

    def __init__(self, name, dtype=float, obj_ref=None, precision=None):
        if not dtype or not inspect.isclass(dtype) or not ott.is_type_supported(dtype):
            raise TypeError(f'Column does not support "{dtype}" type')

        # validating column name
        if name != 'Time':  # this is a special value
            self._check_name(name)

        self.name = name
        super().__init__(dtype=dtype, obj_ref=obj_ref, op_str=name)

        # optional properties
        if precision is not None:
            if issubclass(dtype, float):
                self._precision = precision
            else:
                raise ValueError("precision is supported only for columns with float or decimal dtypes")

    def rename(self, new_name, update_parent_object=True):
        self._check_name(new_name)
        if self.obj_ref and update_parent_object:
            self.obj_ref.rename({self.name: new_name}, inplace=True)

        self.name = new_name

    def __len__(self):
        if issubclass(self.dtype, str):
            if issubclass(self.dtype, ott.string):
                return self.dtype.length
            else:
                return ott.string.DEFAULT_LENGTH
        else:
            raise TypeError(f'It is not applicable for the column with type {self.dtype}')  # TODO: test

    def __hash__(self):
        return hash(self.name)

    def __str__(self):
        return self.name

    def __repr__(self):
        return f"Column({str(self)}, {self.dtype})"

    def copy(self, obj_ref=None):
        return _Column(self.name, self.dtype, obj_ref)

    def __bool__(self):
        if _Column.emulation_enabled:
            if issubclass(self.dtype, int):
                return (self != 0).__bool__()
            if issubclass(self.dtype, float):
                return (self != 0).__bool__()
            if issubclass(self.dtype, str):
                return (self != "").__bool__()

        raise TypeError("It is not allowed to use columns in if-else and while clauses")

    def __getitem__(self, item):

        """
        Provides an ability to get values from future or past ticks.

        - Negative values refer to past ticks

        - Zero to current tick

        - Positive - future ticks

        Boundary values will be defaulted. For instance for ``item=-1`` first tick value will be defaulted
        (there is no tick before first tick)

        Parameters
        ----------
        item: int
            number of ticks to look back/forward

        Returns
        -------
        Operation

        Examples
        --------
        >>> data = otp.Ticks({'A': [1, 2, 3]})
        >>> data['PAST1'] = data['A'][-1]
        >>> data['PAST2'] = data['A'][-2]
        >>> data['FUTURE1'] = data['A'][1]
        >>> data['FUTURE2'] = data['A'][2]
        >>> otp.run(data)
                             Time  A  PAST1  PAST2  FUTURE1  FUTURE2
        0 2003-12-01 00:00:00.000  1      0      0        2        3
        1 2003-12-01 00:00:00.001  2      1      0        3        0
        2 2003-12-01 00:00:00.002  3      2      1        0        0
        """

        if not isinstance(item, int):
            raise TypeError(
                f"Lag operation supports only integer const values, but passed value of type '{type(item)}'"
            )
        if item == 0:
            return self

        return _LagOperator(self, item)

    def cumsum(self):
        """
        Cumulative sum of the column.

        Can only be used when creating or updating column.

        Examples
        --------
        >>> t = otp.Ticks({'A': [1, 2, 3]})
        >>> t['X'] = t['A'].cumsum()
        >>> otp.run(t)
                             Time  A  X
        0 2003-12-01 00:00:00.000  1  1
        1 2003-12-01 00:00:00.001  2  3
        2 2003-12-01 00:00:00.002  3  6
        """
        import onetick.py as otp

        return _ColumnAggregation(
            otp.agg.sum(self.name, running=True, all_fields=True, overwrite_output_field=True)
        )

    def __iter__(self):
        raise TypeError("It is not allowed to use columns in for-clauses")


class _LagOperator(_Operation):
    """
    Implements referencing to the prior tick
    """

    def __init__(self, base_column, inx):
        self._inx = inx
        op_str = f"{str(base_column)}[{self.index}]"
        super().__init__(op_params=[base_column], dtype=base_column.dtype,
                         op_str=op_str, obj_ref=base_column.obj_ref)

    @property
    def index(self):
        return self._inx


class _ColumnAggregation:
    """
    Object to specify how column will be aggregated.
    """
    def __init__(self, aggregation):
        from ..aggregations._base import _Aggregation
        if not isinstance(aggregation, _Aggregation):
            raise ValueError(f'Expected aggregation object, got {type(aggregation)}')
        if not aggregation.running or not aggregation.all_fields or not aggregation.overwrite_output_field:
            raise ValueError("Column aggregations only support 'running' aggregations"
                             " with 'all_fields' and 'overwrite_output_field' parameters set")
        self.aggregation = aggregation

    def apply(self, src, name):
        return self.aggregation.apply(src, name=name, inplace=True)


_Column = Column  # alias for backward compatibility
